import pyav.src.cmd_modules.Rbk2IfxTx2Rx16 as Rbk2IfxTx2Rx16
'''
    Designed for MIMO RadarBook2
'''


import  numpy as np
import numpy.matlib

import copy, sys, time, dill, threading

from threading                              import Thread
from tqdm                                   import tqdm 
from scipy                                  import signal

from pyav.SignalProcessing_class            import SignalProcessing
from datetime                               import datetime
# from numba                                  import jit


c0 = 299792458

class RadarBook2():
    def __init__(self,  RMin, 
                        RMax,
                        VMin,
                        VMax,
                        nPointsFFTrange,
                        nPointsFFTvel, 
                        nPointsFFTang, 
                        DelayProcessing,
                        Nchirp = 100 ):

        self.RMin = RMin
        self.RMax = RMax

        self.VMin = VMin
        self.VMax = VMax

        self.NFFT       = nPointsFFTrange
        self.NFFTVel    = nPointsFFTvel
        self.NFFTAnt    = nPointsFFTang

        self.DelayProcessing = DelayProcessing


        
        # Setup Connection
        #--------------------------------------------------------------------------
        #Brd     =   Rbk2IfxTx2Rx16.Rbk2IfxTx2Rx16('RadServe', '127.0.0.1', 8000, '192.168.1.1')
        self.Brd     =   Rbk2IfxTx2Rx16.Rbk2IfxTx2Rx16('PNet','192.168.1.1')

        self.Brd.BrdRst()
        self.Brd.BrdPwrEna()

        # Use 40 MHz clock for Range Doppler processing. The standard 20 MHz can
        # cause problems with the synchronisation of the sampling and the ramps
        self.Brd.BrdSetRole('Ms', 40e6)


        # Load Calibration Data
        #--------------------------------------------------------------------------
        self.dCalCfg = dict()
        self.dCalCfg['Mask'] = 1
        self.dCalCfg['Len'] = 32
        self.CalData = self.Brd.BrdGetCalData(self.dCalCfg)
        self.CalData_abs = abs(self.CalData)
        self.CalData_angle = np.angle(self.CalData)*180/np.pi #express in deg no rad

        # Configure Up-Chirp
        #--------------------------------------------------------------------------
        self.dCfg = dict()
        self.dCfg['fStrt'] = 76e9
        self.dCfg['fStop'] = 77e9
        self.dCfg['TRampUp'] = 102.4e-6
        self.dCfg['TRampDo'] = 18.8e-6
        self.dCfg['N'] = 512    # number of samples per chirp
        # self.dCfg['NrFrms'] = 100000 # different by number of chirp(=ramp frequency)
        self.dCfg['NLoop'] = Nchirp
        self.dCfg['IniEve'] = 0
        # self.dCfg['IniTim'] = 100e-3
        # self.dCfg['CfgTim'] = 10e-6
        

        self.chirp_totalTime = self.dCfg['NLoop'] * (self.dCfg['TRampUp']+self.dCfg['TRampDo'])
        self.dCfg['TInt'] = self.DelayProcessing  # period of each ramp sequence(=NLoop ramp frequency) = period between adjacent packets of (NLoop)chirps

        #--------------------------------------------------------------------------
        # Configure DMA Transfer to copy Cfg.NLoop frames simultaneously.
        # Required to achiev maximal data transfer between FPGA and Soc 
        self.Brd.Set('DmaMult', self.dCfg['NLoop'])
        # Copy only the data of a single channel. RX1. The data of the residual channels
        # is sampled but not transfered
        # self.Brd.Set('NrChn',16)



        #--------------------------------------------------------------------------
        # Configure Antenna activation sequence and timing with RCC1010 being the
        # master; This is the more advanced measurement mode and Cfg.NLoop can be
        # used to automatically repeat the measurements:
        # (1) The FPGA starts the programmed sequence after Cfg.TInt.
        # (2) The programmed sequence (caRampCfg) has to be shorter than TInt
        #--------------------------------------------------------------------------
        self.lRampCfg = list()
        # Up Chirp and trigger sampling
        self.dRampCfg = dict()
        self.dRampCfg['fStrt'] = self.dCfg['fStrt']
        self.dRampCfg['fStop'] = self.dCfg['fStop']
        self.dRampCfg['TRamp'] = self.dCfg['TRampUp']
        self.dRampCfg['TWait'] = 0
        self.dRampCfg['RampCfg'] = self.Brd.SampSeq + self.Brd.PaCtrl_Tx1

        self.lRampCfg.append(self.dRampCfg)
        # Downchirp with no sampling
        self.dRampCfg = dict()
        self.dRampCfg['fStrt'] = self.dCfg['fStop']
        self.dRampCfg['fStop'] = self.dCfg['fStrt']
        self.dRampCfg['TRamp'] = self.dCfg['TRampDo']
        self.dRampCfg['RampCfg'] = self.Brd.PaCtrl_Tx1
        self.dRampCfg['TWait'] = 0
        self.lRampCfg.append(self.dRampCfg)

        self.Tp = self.lRampCfg[0]['TRamp'] + self.lRampCfg[1]['TRamp']  

        self.Brd.RfMeas('RccMs', self.lRampCfg, self.dCfg)

        self.fs = self.Brd.Get('fs')
        self.N = int(self.Brd.Get('N'))
        self.Np = int(self.dCfg['NLoop'])
        self.NrChn = int(self.Brd.Get('NrChn'))
        self.fAdc  = int(self.Brd.Get('fAdc'))
        self.Rcic  = int(self.Brd.Get('CicR'))
        self.B               =   self.dCfg['fStop']-self.dCfg['fStrt'] # bandwitdh
        self.delta_R         =   c0/(2*self.B) # range resolution
        self.delta_v         =   (c0/((self.dCfg['fStop'] + self.dCfg['fStrt'])/2))/(2*self.dCfg['NLoop']*self.Tp) # velocity resolution 
        self.delta_theta     =   2/self.NrChn # angular resolution
        self.Txpos           =   self.Brd.RfGet('TxPosn') #position of the transmit antennas. The transmit antennas are spaced of (lambda*15)/2 each other
        self.Rxpos           =   self.Brd.RfGet('RxPosn') #position of the receive antennas. The receive antennas are spaced of lambda/2 each other
        # TX2 is spaced from RX1 by (lambda*16)/5
        self.f0              =   (self.dCfg['fStrt'] +self.dCfg['fStop'])/2
        self.waveLenght      =   c0/self.f0


        # Configure Signal Processing
        #--------------------------------------------------------------------------
        # Processing of range profile
        self.Win2D = self.Brd.hanning(self.N-1,self.Np)
        self.Win2D_ang = self.Brd.hanning(self.N-1,self.NrChn)

        self.ScaWin = sum(self.Win2D[:,0])
        self.ScaWin_ang = sum(self.Win2D_ang[:,0])


        self.kf = (self.dCfg['fStop'] - self.dCfg['fStrt'])/self.dCfg['TRampUp'] # = B/TrampUp
        self.vRange = np.arange(self.NFFT//2)/self.NFFT*self.fs*c0/(2*self.kf)
        self.fc = (self.dCfg['fStop'] + self.dCfg['fStrt'])/2


        self.RMinIdx = np.argmin(np.abs(self.vRange - self.RMin))
        self.RMaxIdx = np.argmin(np.abs(self.vRange - self.RMax))
        self.vRangeExt = self.vRange[self.RMinIdx:self.RMaxIdx+1 ]


        self.WinVel2D = self.Brd.hanning(int(self.Np), len(self.vRangeExt))
        self.ScaWinVel = sum(self.WinVel2D[:,0])
        self.WinVel2D = self.WinVel2D.T



        self.vFreqVel = np.arange(-self.NFFTVel//2,self.NFFTVel//2)/self.NFFTVel*(1/self.Tp)
        self.vFreqVel_max    =   1/(2*self.Tp)
        self.vFreqVel_step   =   1/(self.NFFTVel*self.Tp) # frequency bin's resolution
        self.vVel            =   self.vFreqVel*c0/(2*self.fc)
        self.vVel_max        =   self.vFreqVel_max*c0/(2.*self.fc) # max velocity = lambda/(4*Tp)
        self.vVel_step       =   self.vFreqVel_step*c0/(2.*self.fc) # velocity bin's resolution

        self.VMinIdx = np.argmin(np.abs(self.vVel - self.VMin))
        self.VMaxIdx = np.argmin(np.abs(self.vVel - self.VMax))
        self.vVelExt = self.vVel[self.VMinIdx:self.VMaxIdx+1]


        # WinAnt2D = self.Brd.hanning(NrChn, len(vRangeExt))
        self.WinAnt2D = self.Brd.hanning(self.NrChn, 1)
        self.ScaWinAnt = np.sum(self.WinAnt2D[:,0])
        self.WinAnt2D = self.WinAnt2D.T
        self.vAngDeg  = np.arcsin(2*np.arange(-self.NFFTAnt//2, self.NFFTAnt//2)/self.NFFTAnt)/np.pi*180
        self.vAngDeg_min = self.vAngDeg[0]
        self.vAngDeg_max = self.vAngDeg[len(self.vAngDeg)-1]

        self.mCalData = np.matlib.repmat(self.CalData, self.N-1,1) # calibration matrix where each column is a repetition of the calibration coefficients of the respective channel. (N-1) copies of CalData' 


        #--------------------------------------------------------------------------
        # Measure and calculate Range Doppler Map
        #--------------------------------------------------------------------------
        # Select channel to be processed. ChnSel <= NrChn
        # ChnSel = 0
        self.mCalData_rep=np.zeros((self.N-1, self.Np, self.NrChn),dtype="complex")
        self.Win2D_ang_rep=np.zeros((self.N-1, self.Np, self.NrChn),dtype="complex")
        self.WinVel2D_rep=np.zeros((self.vRangeExt.size, self.Np, self.NrChn),dtype="complex")

        for j in range( self.mCalData.shape[1]):
            for i in range( self.Np):
                self.mCalData_rep[:,i,j] = copy.copy(self.mCalData[:,j])
                self.Win2D_ang_rep[:,i,j] = copy.copy(self.Win2D_ang[:,j])
            self.WinVel2D_rep[:,:,j] = copy.copy(self.WinVel2D)



        #   ----------------------------------------------------------------------

        #   Print radar info
        self.print_info()

        self.SignalProcessing = SignalProcessing()

        
        #   Saved data
        self.SavedData  = { 'data':[] , 'times':[] , 'acquisitionTime': None, 
                            'emptyroom':{'RD':None, 'RDA':None},
                            'localization':{'absLocalization':[], 'RDALocalization':[]}
                            }

        
        # freeze
        self.freezeBrd = None



    def print_info(self):
        print();print()
        print('Range:')
        print('\tMinimum range:\t'+repr(self.RMin) + ' m')
        print('\tMaximum range:\t'+repr(self.RMax) + ' m')

        print()

        print('\tRange:\t\tnumber of bin: '+ repr(self.getRange_bin().shape[0]), end='\t')
        print('Average range bin size\t'+ repr(round( np.diff(self.vRange).mean(axis=0), 4 )) + ' m')

        print('\tvelocity:\tnumber of bin: '+ repr(self.getVelocity_bin().shape[0]), end='\t')
        print('Average velocity bin size\t'+ repr(round( np.diff(self.getVelocity_bin()).mean(axis=0), 4 )) + ' m/s')

        print('\tangle:\t\tnumber of bin: '+ repr(self.getAngle_bin().shape[0]), end='\t' )
        print('Average angle bin size\t'+ repr(round( np.diff(self.getAngle_bin()).mean(axis=0), 4 )) + ' °')

        print()

        self.dCfg['fStrt'] = 76e9
        self.dCfg['fStop'] = 77e9
        print('\tFrequency:')
        print('\t\tfreq start ' + repr(self.dCfg['fStrt'])+ 'Hz' )
        print('\t\tfreq stop ' + repr(self.dCfg['fStop'])+ ' Hz' )

        print()

        print('\tChirp:')
        print('\t\ttime RampUp\t\t' + repr(self.dCfg['TRampUp'])+ ' s' )
        print('\t\ttime RampDo\t\t' + repr(self.dCfg['TRampDo'])+ ' s' )
        print('\t\tNumber of chirp\t\t' + repr(self.dCfg['NLoop']) )
        print('\t\tTime to collect chirps\t' + repr(self.chirp_totalTime)+' s' ) 


        print()
        print('\tProcessing time '+ repr(round(self.DelayProcessing ,6) ) +' s')
        print('\tTotal delay between measurements ( Processing + collect data time) '+ (repr(self.dCfg['TInt']) if self.dCfg['TInt'] > self.chirp_totalTime else repr(self.chirp_totalTime) )+' s' )
        print('\tAcutal time calculation...')
        delta = []
        for i in tqdm(range(10)):
            t0 = time.time()
            radarData = self.get_radar_data()
            t1 = time.time()
            delta.append(t1-t0) 
        delta=np.array(delta).mean()
        print('\tActual time '+ repr(round(delta ,6) ) +' s')
        print('-------------- GO --------------')

        print();print()

    #    S A M P L E   R A T E   ------------------------------------------------------------------------------------------------
    def get_ChirpDuration(self, N=1):
        '''
        Input:
            N           :   number of chirp sends
                int
        Return:
            delta_t     :   the duration of a chirp multiplyed for the number of chirps
                float
        '''
        delta_t = (self.dCfg['TRampUp'] + self.dCfg['TRampDo'])*N
        return delta_t


    #    R A N G E   B I N   ------------------------------------------------------------------------------------------------
    def getRange_bin(self):
        '''
        Return:
            vRangeExt:      values of bins from RMin to RMax (shape: if N bin, then N+1 value)
                np.array
        '''
        return self.vRangeExt
    


    #    V E L O C I T Y   B I N   ------------------------------------------------------------------------------------------------
    def getVelocity_bin(self):
        '''
        Return:
            vVel:           values of bins (shape: if N bin, then N+1 value)
                np.array
        '''
        return self.vVelExt
    


    #    A N G L E S   B I N   ------------------------------------------------------------------------------------------------
    def getAngle_bin(self):
        '''
        Return:
            vAngDeg:        values of bins (shape: if N bin, then N+1 value)
                np.array
        '''
        return self.vAngDeg



    #    S A V E   ------------------------------------------------------------------------------------------------
    def SaveData(   self, 
                    numMeas,     # number of measurements
                    print_ProgressBar = True,
                    ):
        '''
            Saves a certain number of measurements in the variable self.SavedData.
            self.SavedData is a dict with key { 'data':list of np.array, 
                                                'times': list of time info for each getData, 
                                                'acquisitionTime: datetime time         }    

            Input:
                numMeas:                number of data that will be collected
                    int

                print_progressBar:      choice to print or not the progress bar
                    bool(=True)
                
                saveAs_DataCube:        specify if data is saved as a cube
                    bool(=True)
            
            example:    file_to_store = open("Serialized_obj/meas1.dill", "wb")
                        dill.dump( radar, file_to_store)  # file
                        file_to_store.close()
        '''
        
        acquisitionTime_start = datetime.now()

        if print_ProgressBar:
            for i in tqdm(range(numMeas), colour='green'):

                start_time = datetime.now()
                self.SavedData['data'].append( self.get_radar_data( ) )
                end_time = datetime.now()

                self.SavedData['times'].append({ 'start_time': start_time,
                                                'end_time'  : end_time          })


       
        else:
            for i in range(numMeas):

                start_time = datetime.now()
                self.SavedData['data'].append( self.get_radar_data( ) )
                end_time = datetime.now()

                self.SavedData['times'].append({    'start_time': start_time,
                                                    'end_time'  : end_time          })

        acquisitionTime_end = datetime.now()

        self.SavedData['acquisitionTime'] = acquisitionTime_end - acquisitionTime_start
        
        print('measurement avaiable in "SavedData", shape '+ repr(len(self.SavedData['data'])) + '\tacquisition time: ', self.SavedData['acquisitionTime'] )

    

    #    L O A D   ------------------------------------------------------------------------------------------------
    def LoadData( self ):
        '''
            Return  self.SavedData['data'] that is a list of np.array,
                    self.SavedData['times'] that is a list of time info about measurements
                    self.SavedData['acquisitionTime'] that is a datetime object 
                    ...

            example:    file_to_store = open("Serialized_obj/meas1.dill", "wb")
                        radar = dill.load(file_to_store)  # file
                        file_to_store.close()
        '''
        return  np.array(self.SavedData['data']), self.SavedData['times'], self.SavedData['acquisitionTime'], self.SavedData['emptyroom'], self.SavedData['localization']

        

    #    F R E E Z E    ------------------------------------------------------------------------------------------------
    def freeze(self):
        '''
            To freeze the obj in order to save it on binary file
                for example with library  dill (https://dill.readthedocs.io/en/latest/) or 
                                          pickle (https://docs.python.org/3/library/pickle.html)
        '''
        self.Brd.cRadDatSocket = None
        self.Brd.cRadSocket = None





    #    G E T   D A T A    ------------------------------------------------------------------------------------------------
    def get_radar_data(self):
        '''

            Return:
                    Data:   to collect data with the radar (shape: (n_samples, n_chirp,  16) )
                    
        '''
        Data = self.Brd.BrdGetData(self.Np)
        # Reshape
        Data = np.reshape(Data[:,:], (self.N, self.Np, self.NrChn), order='F')

        # Removing chirp index
        Data = Data[1:,:] 
        return Data
    


    #    R A N G E   F F T     ------------------------------------------------------------------------------------------------
    def compute_rangeFFT(self, DataCube):
        rangeFFT=np.fft.fft(DataCube[:,:,:]*self.Win2D_ang_rep*self.mCalData_rep,n=self.NFFT,axis=0)*self.Brd.FuSca/self.ScaWin_ang
        rangeFFT=rangeFFT[self.RMinIdx: self.RMaxIdx+1,:,:]

        return rangeFFT
    


    #    V E L O C I T Y   F F T     ------------------------------------------------------------------------------------------------
    def compute_velFFT(self, rangeFFT, antenna = 0):
        velFFT=np.fft.fft(rangeFFT[:,:,antenna]*self.WinVel2D, n=self.NFFTVel, axis=1)/self.ScaWinVel
        velFFT=np.fft.fftshift(velFFT, axes=1)
        velFFT=velFFT[:,self.VMinIdx:self.VMaxIdx+1]
        return velFFT



    #    A N G L E   F F T     ------------------------------------------------------------------------------------------------
    def compute_angleFFT(self, rangeFFT, idx = 0):
        angleFFT = np.fft.fft(rangeFFT[ : ,idx,:] * self.WinAnt2D, n=self.NFFTAnt, axis=1)/self.ScaWinAnt
        angleFFT = np.fft.fftshift(angleFFT, axes=1) 

        return angleFFT



    #    R A N G E - D O P P L E R     ------------------------------------------------------------------------------------------------
    def createMap_RangeDoppler(self, DataCube, antenna=0):
        '''
            Create a range dopplere map
            Input:
                DataCube        : Data obtained from  self.get_radar_data() 
                    np.array      shape(n_samples, n_chirp,  16)
                
                antenna         : define the antenna index to use for creating the range doppler map.
                    int/'all'     if 'all' the mean over all antennas will be computed.

            Return:
                rangeDoppler    : Range doppler map
                    np.array      shape(n_samples range FFT from RMin to RMax, n_samples velocity FFT)

        '''

        if antenna == 'all':
            rangeDoppler =np.array([ abs(self.compute_velFFT(self.compute_rangeFFT(DataCube), antenna)).T for antenna in range(16) ]).mean(axis=0)

        else:
            rangeFFT      = self.compute_rangeFFT(DataCube)
            rangeDoppler  = self.compute_velFFT(rangeFFT, antenna)
            rangeDoppler = abs(rangeDoppler).T

        return rangeDoppler



    #    R A N G E - A N G L E     ------------------------------------------------------------------------------------------------
    def createMap_RangeAngle(self, DataCube, chirpIdx=0):
        '''
            Create a range dopplere map
            Input:
                DataCube        : Data obtained from  self.get_radar_data() 
                    np.array      shape(n_samples, n_chirp,  16)
                
                chirpIdx        : define the chirp index to use for creating the range angle map.
                    int/'all'     if 'all' the mean over all chirps will be computed.

            Return:
                rangeAngle      :   range angle map
                    np.array        shape(n_samples range FFT from RMin to RMax, n_samples angles FFT)

        '''
        
        if chirpIdx == 'all':
            rangeAngle =np.array([ abs(self.compute_angleFFT(self.compute_rangeFFT(DataCube), chirpIdx)).T  for chirpIdx in range(DataCube.shape[1])] ).mean(axis=0)

        else:
            rangeFFT = self.compute_rangeFFT(DataCube)
            angleFFT = self.compute_angleFFT(rangeFFT, chirpIdx)
            rangeAngle =abs(angleFFT).T

        return rangeAngle


    #    R A N G E - D O P P L E R - A N G L E     ------------------------------------------------------------------------------------------------
    def createMap_RangeVelocityAngle(self, DataCube ):
        '''
            Create a range dopplere angle map
            Input:
                DataCube        : Data obtained from  self.get_radar_data() 
                    np.array      shape(n_samples, n_chirp,  16)
                
            Return:
                rangeAngle      :   range velocity angle map
                    np.array        shape(n_samples range FFT from RMin to RMax, n_samples vel FFT from VMin to VMax n_samples angles FFT)

        '''

        data = DataCube*self.Win2D_ang_rep*self.mCalData_rep
        R = np.fft.fft(data,n=self.NFFT,axis=0) * self.Brd.FuSca/self.ScaWin_ang
        R= R[ self.RMinIdx: self.RMaxIdx+1, :, :]

        R_temp = np.zeros(R.shape, dtype=complex)
        for antenna in range(16):
            R_temp[:,:,antenna] = R[:,:, antenna] *self.WinVel2D
        R =R_temp
        
        RV = np.fft.fft(R, n=self.NFFTVel,axis=1)/self.ScaWinVel
        RV=np.fft.fftshift(RV, axes=1)
        RV = RV[:,self.VMinIdx:self.VMaxIdx+1,:]

        RV_temp = np.zeros(RV.shape,dtype=complex)
        for idx in range(RV.shape[1]): 
            RV_temp[ : ,idx,:]  = RV[ : ,idx,:] * self.WinAnt2D
        RV = RV_temp

        RVA = np.fft.fft(RV,n=self.NFFTAnt,axis=2)
        RVA=np.fft.fftshift(RVA, axes=2)

        absRVA = abs(RVA)

        return absRVA
    

    #    R A N G E - D O P P L E R   &   R A N G E - D O P P L E R - A N G L E     ------------------------------------------------------------------------------------------------
    def createMAP_RDA_RD(self, DataCube ):
        '''
            Create a range dopplere angle map
            Input:
                DataCube        : Data obtained from  self.get_radar_data() 
                    np.array      shape(n_samples, n_chirp,  16)
                
            Return:
                rangeAngle      :   range velocity angle map
                    np.array        shape(n_samples range FFT from RMin to RMax, n_samples vel FFT from VMin to VMax n_samples angles FFT)

        '''

        data = DataCube*self.Win2D_ang_rep*self.mCalData_rep
        R = np.fft.fft(data,n=self.NFFT,axis=0) * self.Brd.FuSca/self.ScaWin_ang
        R= R[ self.RMinIdx: self.RMaxIdx+1, :, :]

        R_temp = np.zeros(R.shape, dtype=complex)
        for antenna in range(16):
            R_temp[:,:,antenna] = R[:,:, antenna] *self.WinVel2D
        R =R_temp
        
        RD = np.fft.fft(R, n=self.NFFTVel,axis=1)/self.ScaWinVel
        RD=np.fft.fftshift(RD, axes=1)
        RD = RD[:,self.VMinIdx:self.VMaxIdx+1,:]

        rangeDoppler = abs(RD).mean(axis=2).T
        
        RD_temp = np.zeros(RD.shape,dtype=complex)
        for idx in range(RD.shape[1]): 
            RD_temp[ : ,idx,:]  = RD[ : ,idx,:] * self.WinAnt2D
        RD = RD_temp

        RDA = np.fft.fft(RD,n=self.NFFTAnt,axis=2)
        RDA=np.fft.fftshift(RDA, axes=2)

        rangeDopplerAngle = abs(RDA)

        return rangeDopplerAngle, rangeDoppler


    #    R D   &   R D A   &   R bins A   f o r   c l a s s i f i c a t i o n    ------------------------------------------------------------------------------------------------
    def createMAP_RD_RDA_RBA(self, DataCube ):
        '''
            Create a range dopplere angle map
            Input:
                DataCube        : Data obtained from  self.get_radar_data() 
                    np.array      shape(n_samples, n_chirp,  16)
                
            Return:
                rangeDopplerAngle      :   RDA reduced (i.e. relevant velocity bins)
                    np.array                
                
                rangeDoppler            :   RD reduced (i.e. relevant velocity bins)
                    np.array  

                rangeBinsAngle          :   RA for each bins 
                    np.array  
        '''


        data = DataCube*self.Win2D_ang_rep*self.mCalData_rep
        R = np.fft.fft(data,n=self.NFFT,axis=0) * self.Brd.FuSca/self.ScaWin_ang
        R= R[ self.RMinIdx: self.RMaxIdx+1, :, :]

        R_temp = np.zeros(R.shape, dtype=complex)
        for antenna in range(16):
            R_temp[:,:,antenna] = R[:,:, antenna] *self.WinVel2D
        # R =R_temp
        
        RD = np.fft.fft(R_temp, n=self.NFFTVel,axis=1)/self.ScaWinVel
        RD=np.fft.fftshift(RD, axes=1)
        RD = RD[:,self.VMinIdx:self.VMaxIdx+1,:]

        rangeDoppler = abs(RD).mean(axis=2).T
        
        RD_temp = np.zeros(RD.shape,dtype=complex)
        for idx in range(RD.shape[1]): 
            RD_temp[ : ,idx,:]  = RD[ : ,idx,:] * self.WinAnt2D
        # RD = RD_temp

        RDA = np.fft.fft(RD_temp,n=self.NFFTAnt,axis=2)
        RDA=np.fft.fftshift(RDA, axes=2)

        rangeDopplerAngle = abs(RDA)
        
        #   R A       ------------------------------------
        R_temp = np.zeros(R.shape,dtype=complex)
        for idx in range(R.shape[1]): 
            R_temp[ : ,idx,:]  = R[ : ,idx,:] * self.WinAnt2D
        RbA = np.fft.fft(R_temp,n=self.NFFTAnt,axis=2)
        RbA=np.fft.fftshift(RbA, axes=2)
        rangeBinAngle = abs(RbA)

        return rangeDopplerAngle, rangeDoppler, rangeBinAngle
        #              RDA              RD          RbA=range-bins-angle for classification


    #   G E T   A N G L E   G I V E N   A   R A N G E   B I N     ------------------------------------------------------------------------------------------------
    def getAngle_givenRangeBin(self, centroid, dataCube, noise=None ):
        '''
            Given a range bin return the most likelihood angle. Use the angle FFT peak.

            Input:
                rangeFFT        :   the output of self.compute_rangeFFT( ... )
                    3D np.array

                rangeBinIndex   :   the identified range bin for which the angle is calculated
                    int
                
                noise       :   background noise
                    2D np.array


            Return:
                angle           :   is the closest angle bin to the FFT peak angle.
                    angle [deg]
        '''

        r , v = centroid

        rIdx = np.abs( self.getRange_bin() - r ).argmin()

        if noise is not None    : angleFFTvector = self.createMap_RangeAngle(dataCube, 0)[:,rIdx] - noise[:,rIdx]
        else                    : angleFFTvector = self.createMap_RangeAngle(dataCube, 0)

        angle = self.getAngle_bin()[ angleFFTvector.argmax() ]
        return angle


    
    # #   N U M B A   R A N G E - D O P P L E R    M A P     ------------------------------------------------------------------------------------------------
    # @jit
    # def createRD_withNumba(self, radarData):
    #     '''
    #         Numba is emplyed to make fast the code. The range-Doppler map is created meaning over all the antennas
    #     '''

    #     rangeFFT=np.fft.fft(radarData[:,:,:]*self.Win2D_ang_rep*self.mCalData_rep,n=self.NFFT,axis=0)*self.Brd.FuSca/self.ScaWin_ang
    #     rangeFFT=rangeFFT[self.RMinIdx: self.RMaxIdx+1,:,:]

    #     rangeDoppler = []
    #     for antenna in range(16):
    #         velFFT=np.fft.fft(rangeFFT[:,:,antenna]*self.WinVel2D, n=self.NFFTVel, axis=1)/self.ScaWinVel
    #         velFFT=np.fft.fftshift(velFFT, axes=1)
    #         velFFT=velFFT[:,self.VMinIdx:self.VMaxIdx+1]
    #         rangeDoppler.append(abs(velFFT).T)
        
    #     rangeDoppler = np.array(rangeDoppler ).mean(axis=0)
    #     return rangeDoppler


    # #   N U M B A   R A N G E - A N G L E   M A P     ------------------------------------------------------------------------------------------------
    # @jit
    # def createRA_withNumba(self, radarData):
    #     '''
    #         Return a list of chirp, each represented with its range-angle map
    #     '''
    #     allChirp = []

    #     rangeFFT=np.fft.fft(radarData[:,:,:]*self.Win2D_ang_rep*self.mCalData_rep,n=self.NFFT,axis=0)*self.Brd.FuSca/self.ScaWin_ang
    #     rangeFFT=rangeFFT[self.RMinIdx: self.RMaxIdx+1,:,:]

    #     for idx in range(radarData.shape[1]):
    #         angleFFT = np.fft.fft(rangeFFT[ : ,idx,:] * self.WinAnt2D, n=self.NFFTAnt, axis=1)/self.ScaWinAnt
    #         angleFFT = abs(np.fft.fftshift(angleFFT, axes=1) ).T
    #         allChirp.append( angleFFT )

    #     rangeAngle = np.array(allChirp ).mean(axis=0)
    #     return rangeAngle

    # '''


    ####    M O R E    ############################################################################################
    def debug(self):
        '''
            Debug function inside class
        '''
        d = 0

    
    


    

    #    R D   &   R D A   &   R bins A   f o r   c l a s s i f i c a t i o n    ------------------------------------------------------------------------------------------------
    def createMAP_RD_RDA_RBA_threading(self, DataCube, numThreads=15 ):
        '''
            Create a range dopplere angle map
            Input:
                DataCube        : Data obtained from  self.get_radar_data() 
                    np.array      shape(n_samples, n_chirp,  16)

                numThreads      : Number of threads used
                    int
                
            Return:
                rangeDopplerAngle      :   RDA reduced (i.e. relevant velocity bins)
                    np.array                
                
                rangeDoppler            :   RD reduced (i.e. relevant velocity bins)
                    np.array  

                rangeBinsAngle          :   RA for each bins 
                    np.array  
        '''
        def fft_thread(index, **kwargs):
            data_chunks[index] = np.fft.fft(data_chunks[index], **kwargs)


        data = DataCube*self.Win2D_ang_rep*self.mCalData_rep

        # R = np.fft.fft(data,n=self.NFFT,axis=0) * self.Brd.FuSca/self.ScaWin_ang
        data_chunks = np.array_split(data, numThreads, axis=1) # axis 1 for because fft is on axis=0
        # threads = [ threading.Thread( target=np.fft.fft, kwargs={'a':data_chunks[i], 'n':self.NFFT, 'axis': 0} )     for i in range(len(data_chunks))]
        threads = [ threading.Thread( target=fft_thread, args=(i, ), kwargs={'n':self.NFFT, 'axis': 0} )     for i in range(len(data_chunks))]
        for t in threads: t.start()
        for t in threads: t.join()
        R = np.concatenate(data_chunks, axis=1) * self.Brd.FuSca/self.ScaWin_ang
        R= R[ self.RMinIdx: self.RMaxIdx+1, :, :]


        R_temp = np.zeros(R.shape, dtype=complex)
        for antenna in range(16):
            R_temp[:,:,antenna] = R[:,:, antenna] *self.WinVel2D

        
        # RD = np.fft.fft(R_temp, n=self.NFFTVel,axis=1)/self.ScaWinVel
        data_chunks = np.array_split(R_temp, numThreads, axis=0) # axis 0 for because fft is on axis=1
        # threads = [ threading.Thread( target=np.fft.fft, kwargs={'a':data_chunks[i], 'n':self.NFFTVel, 'axis': 1} )     for i in range(len(data_chunks))]
        threads = [ threading.Thread( target=fft_thread, args=(i, ), kwargs={'n':self.NFFTVel, 'axis': 1} )     for i in range(len(data_chunks))]
        for t in threads: t.start()
        for t in threads: t.join()
        RD = np.concatenate(data_chunks, axis=0)/self.ScaWinVel
        RD=np.fft.fftshift(RD, axes=1)
        RD = RD[:,self.VMinIdx:self.VMaxIdx+1,:]

        rangeDoppler = abs(RD).mean(axis=2).T
        
        RD_temp = np.zeros(RD.shape,dtype=complex)
        for idx in range(RD.shape[1]): 
            RD_temp[ : ,idx,:]  = RD[ : ,idx,:] * self.WinAnt2D
        # RD = RD_temp

        # RDA = np.fft.fft(RD_temp,n=self.NFFTAnt,axis=2)
        data_chunks = np.array_split(RD_temp, numThreads, axis=0) # axis 0 for because fft is on axis=2
        # threads = [ threading.Thread( target=np.fft.fft, kwargs={'a':data_chunks[i], 'n':self.NFFTAnt, 'axis': 2} )     for i in range(len(data_chunks))]
        threads = [ threading.Thread( target=fft_thread, args=(i, ), kwargs={'n':self.NFFTAnt, 'axis': 2} )     for i in range(len(data_chunks))]
        for t in threads: t.start()
        for t in threads: t.join()
        RDA = np.concatenate(data_chunks, axis=0)
        RDA=np.fft.fftshift(RDA, axes=2)

        rangeDopplerAngle = abs(RDA)
        
        #   R A       ------------------------------------
        R_temp = np.zeros(R.shape,dtype=complex)
        for idx in range(R.shape[1]): 
            R_temp[ : ,idx,:]  = R[ : ,idx,:] * self.WinAnt2D

        # RbA = np.fft.fft(R_temp,n=self.NFFTAnt,axis=2)
        data_chunks = np.array_split(R_temp, numThreads, axis=0) # axis 0 for because fft is on axis=2
        # threads = [ threading.Thread( target=np.fft.fft, kwargs={'a':data_chunks[i], 'n':self.NFFTAnt, 'axis': 2} )     for i in range(len(data_chunks))]
        threads = [ threading.Thread( target=fft_thread, args=(i, ), kwargs={'n':self.NFFTAnt, 'axis': 2} )     for i in range(len(data_chunks))]
        for t in threads: t.start()
        for t in threads: t.join()
        RbA = np.concatenate(data_chunks, axis=0)
        RbA=np.fft.fftshift(RbA, axes=2)
        rangeBinAngle = abs(RbA)

        return rangeDopplerAngle, rangeDoppler, rangeBinAngle
        #              RDA              RD          RbA=range-bins-angle for classification


"""This module contains the classes and methods for the remote controll of the
RadarLog base-band board
   """
import  os
import  sys
import  ctypes
import  socket
from    numpy import *
import  pyav.src.cmd_modules.Rbk2 as Rbk2
import  pyav.src.cmd_modules.DevRcc1010  as DevRcc1010
import  pyav.src.cmd_modules.DevRrn7745  as DevRrn7745
import  pyav.src.cmd_modules.SeqTrig     as SeqTrig 
import  weakref
import  traceback

#< @version  2.3.0
#< Support GPS mode in RccMs Mode

class Rbk2IfxTx2Rx16(Rbk2.Rbk2):

    ## Constructor
    def __init__(self, stConType, *args):
        super(Rbk2IfxTx2Rx16, self).__init__(stConType, *args)

        #------------------------------------------------------------------
        # Define flags of RRN7745 registers
        #------------------------------------------------------------------          
        self.RRN7745_REG_ENA_MIX_DAC_TST = 1
        self.RRN7745_REG_ENA_MIX_DAC_LO = 2
        self.RRN7745_REG_ENA_MIX_RX_1 = 4
        self.RRN7745_REG_ENA_MIX_RX_2 = 8
        self.RRN7745_REG_ENA_MIX_RX_3 = 16
        self.RRN7745_REG_ENA_MIX_RX_4 = 32
        self.RRN7745_REG_ENA_MIX_RX_ALL = 4 + 8 + 16 + 32
        self.RRN7745_REG_ENA_MIX_RST = 64 + 128
        self.RRN7745_REG_MUX_TRI = 0
        self.RRN7745_REG_MUX_TEMP = 16
        self.RRN7745_REG_MUX_PTST = 32
        self.RRN7745_REG_MUX_PLO = 48
        self.RRN7745_REG_DCX_VAL_OFFS = 32   
    
        #------------------------------------------------------------------
        # Define flags of Rpn7720 registers
        #------------------------------------------------------------------        
        self.RPN7720_REG_PACON_MUX_P1 = 3
        self.RPN7720_REG_PACON_MUX_P2 = 3*16   
        
        self.PaCtrl_TxOff = 0        
        self.PaCtrl_Tx1 = 2
        self.PaCtrl_Tx2 = 4
        self.TxPwr = 63
        
        # P5 is used to control the sampling chain
        self.SampSeq = int('000001', 2)   

        self.Rf_fStrt = 76e9
        self.Rf_fStop = 77e9
        self.Rf_TRampUp = 256e-6
        self.Rf_TRampDo = 256e-6
        
        self.Rf_Rpn7720Cfg1_Mask = 1
        self.Rf_Rpn7720Cfg1_LoChain = int('01111111',2)
        self.Rf_Rpn7720Cfg1_Pa1 = 0
        self.Rf_Rpn7720Cfg1_Pa2 = 0
        self.Rf_Rpn7720Cfg1_PaCon = 0
        self.Rf_Rpn7720Cfg1_Mux = int('10000',2)
        
        #--------------------------------------------------------------------------
        # Static RX Rrn7745 configuration
        #--------------------------------------------------------------------------
        # RX 1
        self.Rf_Rrn7745_Cfg1_Mask = 1
        self.Rf_Rrn7745_Cfg1_Tst_I = int('10000000', 2)
        self.Rf_Rrn7745_Cfg1_Tst_Q = int('10000000', 2)
        self.Rf_Rrn7745_Cfg1_Lo_I = int('00000000', 2)
        self.Rf_Rrn7745_Cfg1_Lo_Q = int('10000000', 2)
        self.Rf_Rrn7745_Cfg1_Ena_Mix = 2 + 4 + 8 + 16 + 32 + 64 + 128
        self.Rf_Rrn7745_Cfg1_Mux = 0
        self.Rf_Rrn7745_Cfg1_Dc1 = 4*32
        self.Rf_Rrn7745_Cfg1_Dc2 = 4*32
        self.Rf_Rrn7745_Cfg1_Dc3 = 4*32
        self.Rf_Rrn7745_Cfg1_Dc4 = 4*32
        
        # Rx 2
        self.Rf_Rrn7745_Cfg2_Mask = 2
        self.Rf_Rrn7745_Cfg2_Tst_I = int('10000000', 2)
        self.Rf_Rrn7745_Cfg2_Tst_Q = int('10000000', 2)
        self.Rf_Rrn7745_Cfg2_Lo_I = int('00000000', 2)
        self.Rf_Rrn7745_Cfg2_Lo_Q = int('10000000', 2)
        self.Rf_Rrn7745_Cfg2_Ena_Mix = 2 + 4 + 8 + 16 + 32 + 64 + 128
        self.Rf_Rrn7745_Cfg2_Mux = 0
        self.Rf_Rrn7745_Cfg2_Dc1 = 4*32
        self.Rf_Rrn7745_Cfg2_Dc2 = 4*32
        self.Rf_Rrn7745_Cfg2_Dc3 = 4*32
        self.Rf_Rrn7745_Cfg2_Dc4 = 4*32

        # Rx 3
        self.Rf_Rrn7745_Cfg3_Mask = 4
        self.Rf_Rrn7745_Cfg3_Tst_I = int('10000000', 2)
        self.Rf_Rrn7745_Cfg3_Tst_Q = int('10000000', 2)
        self.Rf_Rrn7745_Cfg3_Lo_I = int('00000000', 2)
        self.Rf_Rrn7745_Cfg3_Lo_Q = int('10000000', 2)
        self.Rf_Rrn7745_Cfg3_Ena_Mix = 2 + 4 + 8 + 16 + 32 + 64 + 128
        self.Rf_Rrn7745_Cfg3_Mux = 0
        self.Rf_Rrn7745_Cfg3_Dc1 = 4*32
        self.Rf_Rrn7745_Cfg3_Dc2 = 4*32
        self.Rf_Rrn7745_Cfg3_Dc3 = 4*32
        self.Rf_Rrn7745_Cfg3_Dc4 = 4*32
        
        # Rx 4
        self.Rf_Rrn7745_Cfg4_Mask = 8
        self.Rf_Rrn7745_Cfg4_Tst_I = int('10000000', 2)
        self.Rf_Rrn7745_Cfg4_Tst_Q = int('10000000', 2)
        self.Rf_Rrn7745_Cfg4_Lo_I = int('00000000', 2)
        self.Rf_Rrn7745_Cfg4_Lo_Q = int('10000000', 2)
        self.Rf_Rrn7745_Cfg4_Ena_Mix = 2 + 4 + 8 + 16 + 32 + 64 + 128
        self.Rf_Rrn7745_Cfg4_Mux = 0
        self.Rf_Rrn7745_Cfg4_Dc1 = 4*32
        self.Rf_Rrn7745_Cfg4_Dc2 = 4*32
        self.Rf_Rrn7745_Cfg4_Dc3 = 4*32
        self.Rf_Rrn7745_Cfg4_Dc4 = 4*32        
        
        self.Rf_RccCfg_Mask = 1
        self.stRfVers = '2.3.0'
        self.Rf_Rpn7720_Pwr1 = 63
        self.Rf_Rpn7720_Pwr2 = 63

        self.RCC1010_PaCfg_Pa1 = 1   
        self.RCC1010_PaCfg_Pa2 = 2 
        self.RCC1010_PaCfg_Pa3 = 4
        self.RCC1010_PaCfg_Pa4 = 8
        self.RCC1010_PaCfg_LO = 16
        self.RCC1010_PaCfg_EN = 128 

        self.Rcc1010 = DevRcc1010.DevRcc1010(weakref.ref(self), 1)

        self.Rad_FUsbCfg_WaitLev = 2**15-4096
        self.Set('NrChn',16)
        
        self.Computation.Enable();

    # DOXYGEN ------------------------------------------------------
    #> @brief Get Version information of frontend class
    #>
    #> Get version of class 
    #>      - Version string is returned as string
    #>
    #> @return  Returns the version string of the class (e.g. 0.5.0)          
    def     RfGetVers(self):            
        return self.stRfVers

    # DOXYGEN -------------------------------------------------
    #> @brief Displays version information in Matlab command window  
    #>
    #> Display version of class in Matlab command window            
    def     RfDispVers(self):         
        print('Rbk2IfxTx2Rx16 Class Version: ',self.stRfVers) 

    # DOXYGEN ------------------------------------------------------
    #> @brief Set attribute of class object
    #>
    #> Sets different attributes of the class object
    #>
    #> @param[in]     stSel: String to select attribute
    #> 
    #> @param[in]     Val: value of the attribute (optional); can be a string or a number  
    #>  
    #> Supported parameters  
    #>      -   <span style="color: #ff9900;"> 'TxPwrAll': </span> TxPwr with PaControl <br>
    #>          TxPwr 0..255: Control of TX output power in TxSeq mode    
    #>      -   <span style="color: #ff9900;"> 'Tx1Pwr': </span> TxPwr with PaControl for Tx1 <br>
    #>          TxPwr 0..255: Control of TX output power in TxSeq mode    
    #>      -   <span style="color: #ff9900;"> 'Tx2Pwr': </span> TxPwr with PaControl for Tx2 <br>
    #>          TxPwr 0..255: Control of TX output power in TxSeq mode    
    #>      -   <span style="color: #ff9900;"> 'Tx3Pwr': </span> TxPwr with PaControl for Tx3 <br>
    #>          TxPwr 0..255: Control of TX output power in TxSeq mode    
    #>      -   <span style="color: #ff9900;"> 'Tx4Pwr': </span> TxPwr with PaControl for Tx4 <br>
    #>          TxPwr 0..255: Control of TX output power in TxSeq mode         
    #>      -   <span style="color: #ff9900;"> 'Tx5Pwr': </span> TxPwr with PaControl for Tx5 <br>
    #>          TxPwr 0..255: Control of TX output power in TxSeq mode          
    #>      -   <span style="color: #ff9900;"> 'Tx4Pwr': </span> TxPwr with PaControl for Tx6 <br>
    #>          TxPwr 0..255: Control of TX output power in TxSeq mode          
    def     RfSet(self,*varargin):
        if len(varargin) > 0:
            stVal = varargin[0] 
            if stVal == 'Tx1Pwr':
                self.Rf_Rpn7720_Pwr1 = varargin[1] % 64
            elif stVal == 'Tx2Pwr':
                self.Rf_Rpn7720_Pwr2 = varargin[1] % 64                      
            elif stVal == 'TxPwrAll':
                self.Rf_Rpn7720_Pwr1 = varargin[1] % 64
                self.Rf_Rpn7720_Pwr2 = varargin[1] % 64  

    # DOXYGEN ------------------------------------------------------
    #> @brief Get attribute of class object
    #>
    #> Reads back different attributs of the object
    #>
    #> @param[in]   stSel: String to select attribute
    #> 
    #> @return      Val: value of the attribute (optional); can be a string or a number  
    #>  
    #> Supported parameters  
    #>      -   <span style="color: #ff9900;"> 'TxPosn': </span> Positions of transmit anntennas in m <br>
    #>      -   <span style="color: #ff9900;"> 'RxPosn': </span> Positions of receive antennas in m <br>
    #>      -   <span style="color: #ff9900;"> 'ChnDelay': </span> Channel delay shifts<br>
    #>      -   <span style="color: #ff9900;"> 'B': </span> bandwidth of frequency chirp <br>
    #>      -   <span style="color: #ff9900;"> 'kf': </span>  Slope of frequency ramp
    #>      -   <span style="color: #ff9900;"> 'kfUp': </span> Slope of frequency ramp upchirp <br>
    #>      -   <span style="color: #ff9900;"> 'kfDo': </span> Slope of frequency ramp downchirp <br>
    #>      -   <span style="color: #ff9900;"> 'fc': </span> Center frequency <br>
    #>        
    #> e.g. Read Tx-Positions
    #>   @code
    #>      Brd = Rbk2IfxTx2Rx16( )
    #>      
    #>      Brd.RfGet('TxPosn')
    #>              
    #>   @endcode  
    def     RfGet(self, *varargin):
        if len(varargin) > 0:
            stVal = varargin[0]
            if stVal == 'TxGain':
                Ret = ones(2)*16.5
            elif stVal == 'RxGain':
                Ret = ones(16)*14.2
            elif stVal == 'TxPosn':
                Ret = asarray([-42.114, -12.864])*1e-3
            elif stVal == 'RxPosn':
                Ret = linspace(0,15,16)*1.948e-3
            elif stVal == 'B':
                Ret = self.Rf_fStop - self.Rf_fStrt
            elif stVal == 'kf':
                Ret = (self.Rf_fStop - self.Rf_fStrt)/self.Rf_TRampUp
            elif stVal == 'kfUp':
                Ret = (self.Rf_fStop - self.Rf_fStrt)/self.Rf_TRampUp
            elif stVal == 'kfDo':
                Ret = (self.Rf_fStop - self.Rf_fStrt)/self.Rf_TRampDo
            elif stVal == 'fc':
                Ret = (self.Rf_fStop + self.Rf_fStrt)/2
            else:
                Ret = []
        return Ret
    
    # DOXYGEN ------------------------------------------------------
    #> @brief Enable Receive chips
    #>
    #> Function is not used in L-class; receivers are enabled by
    #> default; configuration of receiver is performed in FPGA
    #>           
    def     RfRxEna(self):
        dCfg = dict()
        dCfg['Mask'] = self.Rf_Rrn7745_Cfg1_Mask
        dCfg['Tst_I'] = self.Rf_Rrn7745_Cfg1_Tst_I
        dCfg['Tst_Q'] = self.Rf_Rrn7745_Cfg1_Tst_Q
        dCfg['Lo_I'] = self.Rf_Rrn7745_Cfg1_Lo_I
        dCfg['Lo_Q'] = self.Rf_Rrn7745_Cfg1_Lo_Q
        dCfg['Ena_Mix'] = self.Rf_Rrn7745_Cfg1_Ena_Mix
        dCfg['Mux'] = self.Rf_Rrn7745_Cfg1_Mux
        dCfg['Dc1'] = self.Rf_Rrn7745_Cfg1_Dc1
        dCfg['Dc2'] = self.Rf_Rrn7745_Cfg1_Dc2
        dCfg['Dc3'] = self.Rf_Rrn7745_Cfg1_Dc3
        dCfg['Dc4'] = self.Rf_Rrn7745_Cfg1_Dc4
        self.RfRrn7745Ini(dCfg)

        dCfg = dict()
        dCfg['Mask'] = self.Rf_Rrn7745_Cfg2_Mask
        dCfg['Tst_I'] = self.Rf_Rrn7745_Cfg2_Tst_I
        dCfg['Tst_Q'] = self.Rf_Rrn7745_Cfg2_Tst_Q
        dCfg['Lo_I'] = self.Rf_Rrn7745_Cfg2_Lo_I
        dCfg['Lo_Q'] = self.Rf_Rrn7745_Cfg2_Lo_Q
        dCfg['Ena_Mix'] = self.Rf_Rrn7745_Cfg2_Ena_Mix
        dCfg['Mux'] = self.Rf_Rrn7745_Cfg2_Mux
        dCfg['Dc1'] = self.Rf_Rrn7745_Cfg2_Dc1
        dCfg['Dc2'] = self.Rf_Rrn7745_Cfg2_Dc2
        dCfg['Dc3'] = self.Rf_Rrn7745_Cfg2_Dc3
        dCfg['Dc4'] = self.Rf_Rrn7745_Cfg2_Dc4
        self.RfRrn7745Ini(dCfg)

        dCfg = dict()
        dCfg['Mask'] = self.Rf_Rrn7745_Cfg3_Mask
        dCfg['Tst_I'] = self.Rf_Rrn7745_Cfg3_Tst_I
        dCfg['Tst_Q'] = self.Rf_Rrn7745_Cfg3_Tst_Q
        dCfg['Lo_I'] = self.Rf_Rrn7745_Cfg3_Lo_I
        dCfg['Lo_Q'] = self.Rf_Rrn7745_Cfg3_Lo_Q
        dCfg['Ena_Mix'] = self.Rf_Rrn7745_Cfg3_Ena_Mix
        dCfg['Mux'] = self.Rf_Rrn7745_Cfg3_Mux
        dCfg['Dc1'] = self.Rf_Rrn7745_Cfg3_Dc1
        dCfg['Dc2'] = self.Rf_Rrn7745_Cfg3_Dc2
        dCfg['Dc3'] = self.Rf_Rrn7745_Cfg3_Dc3
        dCfg['Dc4'] = self.Rf_Rrn7745_Cfg3_Dc4
        self.RfRrn7745Ini(dCfg)    

        dCfg = dict()
        dCfg['Mask'] = self.Rf_Rrn7745_Cfg4_Mask
        dCfg['Tst_I'] = self.Rf_Rrn7745_Cfg4_Tst_I
        dCfg['Tst_Q'] = self.Rf_Rrn7745_Cfg4_Tst_Q
        dCfg['Lo_I'] = self.Rf_Rrn7745_Cfg4_Lo_I
        dCfg['Lo_Q'] = self.Rf_Rrn7745_Cfg4_Lo_Q
        dCfg['Ena_Mix'] = self.Rf_Rrn7745_Cfg4_Ena_Mix
        dCfg['Mux'] = self.Rf_Rrn7745_Cfg4_Mux
        dCfg['Dc1'] = self.Rf_Rrn7745_Cfg4_Dc1
        dCfg['Dc2'] = self.Rf_Rrn7745_Cfg4_Dc2
        dCfg['Dc3'] = self.Rf_Rrn7745_Cfg4_Dc3
        dCfg['Dc4'] = self.Rf_Rrn7745_Cfg4_Dc4
        self.RfRrn7745Ini(dCfg)                

    # DOXYGEN ------------------------------------------------------
    #> @brief Display RF chip status
    #>
    #> Display status of RF chips
    def     BrdDispSts(self):
        Val = self.Fpga_GetRfChipSts(1)

        if len(Val) > 2:
            print('RF UID = ', hex(Val[0]))
            print('RF RevNr = ', self.RevStr(Val[1]))
            print('RF SerialNr = ', hex(Val[2]))
            print('RF Date = ', Val[3])
            print('RF Startup = ', Val[4])
            print('--------------------')
            print('RCC = ',hex(Val[5]), hex(Val[6]))
            print('DPA1 = ',hex(Val[7]), hex(Val[8]))
            print('MRX1 = ',hex(Val[9]), hex(Val[10]))
            print('MRX2 = ',hex(Val[11]), hex(Val[12]))
            print('MRX3 = ',hex(Val[13]), hex(Val[14]))
            print('MRX4 = ',hex(Val[15]), hex(Val[16]))
        
    # DOXYGEN ------------------------------------------------------
    #> @brief Enable transmit antenna static
    #>
    #> Set static setup for tx antenna and power
    #> 
    #> @param[in] TxChn: Transmit antenna to activate 0-6; 0->all off
    #>
    #> @param[in] TxPwr: Output power setting 0..255 (register setting for tranmitter)
    #>                  - 255 maximum output power 
    def     RfTxEna(self, TxChn, TxPwr):
        TxChn = TxChn % 3
        TxPwr = TxPwr % 64
        
        dCfg1 = dict()
        if TxChn == 0:
           dCfg1['Mask'] = self.Rf_Rpn7720Cfg1_Mask
           dCfg1['LoChain'] = self.Rf_Rpn7720Cfg1_LoChain
           dCfg1['Pa1'] = 0
           dCfg1['Pa2'] = 0
           dCfg1['PaCon'] = 0
           dCfg1['Mux'] = self.Rf_Rpn7720Cfg1_Mux                        
        elif TxChn == 1:
           dCfg1['Mask'] = self.Rf_Rpn7720Cfg1_Mask
           dCfg1['LoChain'] = self.Rf_Rpn7720Cfg1_LoChain
           dCfg1['Pa1'] = TxPwr
           dCfg1['Pa2'] = 0
           dCfg1['PaCon'] = self.RPN7720_REG_PACON_MUX_P1
           dCfg1['Mux'] = self.Rf_Rpn7720Cfg1_Mux    
        elif TxChn == 2:
           dCfg1['Mask'] = self.Rf_Rpn7720Cfg1_Mask
           dCfg1['LoChain'] = self.Rf_Rpn7720Cfg1_LoChain
           dCfg1['Pa1'] = 0
           dCfg1['Pa2'] = TxPwr
           dCfg1['PaCon'] = self.RPN7720_REG_PACON_MUX_P2
           dCfg1['Mux'] = self.Rf_Rpn7720Cfg1_Mux
        else:   
           dCfg1['Mask'] = self.Rf_Rpn7720Cfg1_Mask;
           dCfg1['LoChain'] = self.Rf_Rpn7720Cfg1_LoChain;
           dCfg1['Pa1'] = 0;
           dCfg1['Pa2'] = 0;
           dCfg1['PaCon'] = 0;
           dCfg1['Mux'] = self.Rf_Rpn7720Cfg1_Mux; 
                                     
        self.Rf_Rpn7720Cfg1_Mask = dCfg1['Mask'];
        self.Rf_Rpn7720Cfg1_LoChain = dCfg1['LoChain'];
        self.Rf_Rpn7720Cfg1_Pa1 = dCfg1['Pa1'];
        self.Rf_Rpn7720Cfg1_Pa2 = dCfg1['Pa2'];
        self.Rf_Rpn7720Cfg1_PaCon = dCfg1['PaCon'];
        self.Rf_Rpn7720Cfg1_Mux = dCfg1['Mux'];
        self.RfRpn7720Ini(dCfg1)
        
    def Fpga_DispBrdInf(self):          
        print(' ');
        print('-------------------------------------------------------------------')
        print('Radarbook2: Board Information')
        Val = self.Fpga_GetBrdInf()
        if len(Val) > 2:
            Temp = Val[2] - 273;
            print('UID = ',Val[1], Val[0])
            print('Temp = ',Temp, 'deg')
            print('Temp Rf = ',Val[3],' deg')
            print('RF1 Vcc = ',Val[4]/100,' V')
            print('RF1 Icc = ',Val[5]/100,' A')
         
        else:
            print('Board does not respond!')
        print('-------------------------------------------------------------------')
        return Val


    def     RfTxPaEna(self, TxPwr):
        TxPwr = TxPwr % 64
        dCfg1 = dict()
        dCfg1['Mask'] = self.Rf_Rpn7720Cfg1_Mask
        dCfg1['LoChain'] = self.Rf_Rpn7720Cfg1_LoChain
        dCfg1['Pa1'] = TxPwr
        dCfg1['Pa2'] = TxPwr
        dCfg1['PaCon'] = 2*16+1
        dCfg1['Mux'] = self.Rf_Rpn7720Cfg1_Mux
          
        self.Rf_Rpn7720Cfg1_Mask = dCfg1['Mask']
        self.Rf_Rpn7720Cfg1_LoChain = dCfg1['LoChain']
        self.Rf_Rpn7720Cfg1_Pa1 = dCfg1['Pa1']
        self.Rf_Rpn7720Cfg1_Pa2 = dCfg1['Pa2']
        self.Rf_Rpn7720Cfg1_PaCon = dCfg1['PaCon']
        self.Rf_Rpn7720Cfg1_Mux = dCfg1['Mux']
        
        self.RfRpn7720Ini(dCfg1)
                
    # DOXYGEN ------------------------------------------------------
    #> @brief Set Actual Ramp Parameters
    #>
    #> Set static setup for tx antenna and power
    #> 
    #> @param[in] Cfg: Configuration structure
    def     SetActRampParams(self, dCfg):
        nStop = round(dCfg['fStop']/40e6*(2**16))
        nStrt = round(dCfg['fStrt']/40e6*(2**16))
        nStep = round(abs(nStop - nStrt)/(40e6*dCfg['TRampUp'] - 1))
        TRampAct = ceil(abs(nStop - nStrt)/nStep + 1)*25e-9    
        self.Rf_fStrt = nStrt/(2**16)*40e6
        self.Rf_fStop = nStop/(2**16)*40e6
        self.Rf_TRampUp = TRampAct
        nStep = round(abs(nStop - nStrt)/(40e6*dCfg['TRampDo'] - 1))
        TRampAct = ceil(abs(nStop - nStrt)/nStep + 1)*25e-9                           
        self.Rf_TRampDo = TRampAct      
 
    # DOXYGEN ------------------------------------------------------
    #> @brief RfMeasurement modes
    #>
    #> Set timing and ramp configuration for the mode
    #> 
    #> @param[in] stSel: String to select predefined mode
    #>
    #> @param[in] Cfg: Structure to configure measurement mode   
    #>      -   <span style="color: #ff9900;"> 'ExtTrigUp': </span> FMCW measurement mode with uniform timing <br>
    #>          -   dCfg['fStrt']       Start frequency in Hz (mandatory)
    #>          -   dCfg['fStop']       Stop frequency in Hz (mandatory)
    #>          -   dCfg['TRampUp']     Upchirp ramp duration (mandatory)
    #>          -   dCfg['TRampDo']     Downchirp ramp duration (mandatory)
    #>          -   dCfg['NrFrms']      Number of measured Upchirps 
    #>          -   dCfg['N']           Desired number of samples N (mandatory)
    #>          -   dCfg['IniEve']      Enable disable trigger event after start phase: (default 1)
    #>                              If set to 0, measurement is started automatically
    #>                              If set to 1, Fpga_SeqTrigExtEve starts the measurement    
    #>          -   dCfg['IniTim']      Time duration of ini phase in s (default 2ms)
    #>                              Can be helpful to delay measurements if initialization takes longer (e.g large USB data frames)
    #>          -   dCfg['CfgTim']      Duration for reconfiguration of PLL (default 50us)
    #>          -   dCfg['ExtEve']      Wait for external event after measurement frame (default 0)
    #>                              If set to 1, Fpga_SeqTrigExtEve must be called to initiate next measurement
    #>          -   dCfg['Start']       Enable disable start of timing in FPGA
    #>          
    #>      -   <span style="color: #ff9900;"> 'ExtTrigUpNp': </span> FMCW measurement mode with NP uniform chirps (sequence) <br>
    #>          -   dCfg['fStrt']       Start frequency in Hz (mandatory)
    #>          -   dCfg['fStop']       Stop frequency in Hz (mandatory)
    #>          -   dCfg['TRampUp']     Upchirp ramp duration (mandatory)
    #>          -   dCfg['TRampDo']     Downchirp ramp duration (mandatory)
    #>          -   dCfg['NrFrms']      Number of sequences (one sequence consists of Np chirps)
    #>          -   dCfg['N']           Desired number of samples N (mandatory)
    #>          -   dCfg['Np']          Number of subsequent frames Np (mandatory)
    #>          -   dCfg['Tp']          Chirp repetition interval (mandatory)
    #>          -   dCfg['TInt']        Interval between adjacent sequencies      
    #>          -   dCfg['IniEve']      Enable disable trigger event after start phase: (default 1)
    #>                              If set to 0, measurement is started automatically
    #>                              If set to 1, Fpga_SeqTrigExtEve starts the measurement    
    #>          -   dCfg['IniTim']      Time duration of ini phase in s (default 2ms)
    #>                              Can be helpful to delay measurements if initialization takes longer (e.g large USB data frames)
    #>          -   dCfg['CfgTim']      Duration for reconfiguration of PLL (default 50us)
    #>          -   dCfg['ExtEve']      Wait for external event after a sequence (default 0)
    #>                              If set to 1, Fpga_SeqTrigExtEve must be called to initiate next measurement sequence (Np frames)
    #>          -   dCfg['Start']       Enable disable start of timing in FPGA                
    def     RfMeas(self, *varargin):
        if len(varargin) > 1:
            stMod = varargin[0]

            if stMod == 'ExtTrigUp':

                print('Simple Measurement Mode: ExtTrigUp');
                self.RfRst()
                self.Rcc1010.SmartIni('ExtTrigMux2')
                dCfg = varargin[1]
                  
                if not ('fStrt' in dCfg):
                    print('RfMeas: fStrt not specified!')
                    traceback.print_exc()
                if not ('fStop' in dCfg):
                    print('RfMeas: fStop not specified!')
                    traceback.print_exc()
                if not ('TRampUp' in dCfg):
                    print('RfMeas: TRampUp not specified!')
                    traceback.print_exc()
                if not ('TRampDo' in dCfg):
                    print('RfMeas: TRampDo not specified!')
                    traceback.print_exc()
                if not ('N' in dCfg):
                    print('RfMeas: N not specified!')
                    traceback.print_exc()
                if not ('TInt' in dCfg):
                    print('RfMeas: TInt not specified!')
                    traceback.print_exc()

                dCfg = self.ChkMeasCfg(dCfg)
                
                self.SetActRampParams(dCfg)
                # Programm Memory
                lRampCfg = list()
                dRampCfg = dict()
                dRampCfg['fStrt'] = dCfg['fStrt']
                dRampCfg['fStop'] = dCfg['fStop']
                dRampCfg['TRamp'] = dCfg['TRampUp']
                lRampCfg.append(dRampCfg)
                dRampCfg = dict()
                dRampCfg['fStrt'] = dCfg['fStop']
                dRampCfg['fStop'] = dCfg['fStrt']
                dRampCfg['TRamp'] = dCfg['TRampDo']
                lRampCfg.append(dRampCfg)
                
                self.Rcc1010.ProgRampMemory(int('0x1000',0), lRampCfg)
                
                # Execute Processing Cals;
                self.Rcc1010.DevProcCall("SetModRun")
                self.Rcc1010.DevProcCall("InitBasic")
                self.Rcc1010.DevProcCall("OutputStaticFrequency", (dCfg["fStrt"] + dCfg["fStop"])/2)
                self.Rcc1010.DevProcCall("SetTxReg",1, int('11001111',2))
                self.Rcc1010.DevProcCall("SetTxReg",2, int('11001111',2))
                self.Rcc1010.DevProcCall("SetTxReg",3, int('11001111',2))
                self.Rcc1010.DevProcCall("StartCoarseTuneCalibration")
                self.Rcc1010.DevProcCall("StartFQMCalibration")

                # Calculate Sampling Rate
                N = dCfg['N']
                IniCic = 1
                if 'IniCic' in dCfg:
                    if dCfg['IniCic'] == 0:
                        IniCic = 0

                if IniCic > 0:
                    Ts = dCfg['TRampUp']/N
                    fs = 1/Ts
                    fAdc = self.Get('fAdc') 
                    Div = floor(fAdc/fs)
                    if Div < 1:
                        Div = 1
                    print('  fAdc:  ', (fAdc/1e6), ' MHz')
                    print('  CicR:  ', (Div), ' ')
                    print('  TSamp: ', (N/(fAdc/Div)/1e-6), ' us')                      
                    # Configure CIC Filter 
                    self.CfgCicFilt(Div)

                self.Set('N', N)

                self.Fpga_MimoSeqRst()
                self.Fpga_MimoSeqNrEntries(1)

                Regs = self.Rcc1010.GetRegsProcCallExtTrig(int('0x1000',0), 2, 1, 0)
                self.Fpga_MimoSeqSetCrocRegs(0, Regs)
                
                self.Rad_FrmCtrlCfg_RegChnCtrl = self.cFRMCTRL_REG_CH0CTRL_FRMCNTRENA + self.cFRMCTRL_REG_CH0CTRL_WAITENA + self.cFRMCTRL_REG_CH0CTRL_GLOBWAITENA;    
                self.BrdSampIni()

                # Programm Timing to Board: Nios is used to initialize
                # pattern
                IniEve = 0
                if 'IniEve' in dCfg:
                    IniEve = dCfg['IniEve']
                IniTim = 0.2e-3
                if 'IniTim' in dCfg:
                    IniTim = dCfg['IniTim']
                CfgTim = 60e-6
                if 'CfgTim' in dCfg:
                    CfgTim = dCfg['CfgTim']
                if CfgTim < 30e-6:
                    CfgTim = 30e-6
                    print('RCC1010 configuration time set to 30 us')

                MeasEve = 0
                if 'ExtEve' in dCfg:
                    MeasEve = dCfg['ExtEve']
                    if dCfg['ExtEve'] > 0:
                        dCfg['MeasTim']     = dCfg['TRampUp'] + dCfg['TRampDo'] + 60e-6   

                if dCfg['TInt'] < dCfg['TRampUp'] + dCfg['TRampDo'] + CfgTim + 1e-6:
                    dCfg['TInt'] = dCfg['TRampUp'] + dCfg['TRampDo'] + CfgTim + 1e-6
                    print('TInt to short; set to ', (dCfg['TInt']/1e-6), ' us')
                
                dTimCfg = dict()
                dTimCfg['IniEve'] = IniEve                         #   Use ExtEve after Ini phase
                dTimCfg['IniTim'] = IniTim                         #   Duration of Ini phase in us 
                dTimCfg['CfgTim'] = CfgTim                         #   Configuration: Configure RCC for ExtTrig 
                dTimCfg['MeasEve'] = MeasEve                        #   Use ExtEve after meas phase
                dTimCfg['MeasTim'] = dCfg['TInt'] - dTimCfg['CfgTim']      

                # TODO Set Timing 
                self.BrdSetTim_RccM1PW(dTimCfg)
                    
                self.Setfs();
                self.SetFuSca(); 
                self.Computation.SetParam('fStrt',   dCfg['fStrt']);
                self.Computation.SetParam('fStop',   dCfg['fStop']);
                self.Computation.SetParam('TRampUp', dCfg['TRampUp']); 
                
                if 'Strt' in dCfg:
                    if dCfg['Strt'] > 0:
                        self.BrdAccessData()
                        self.BrdSampStrt()
                else:
                    self.BrdAccessData()
                    self.BrdSampStrt()

            elif stMod == 'Cw':
                if self.cDebugInf > 0:
                    print('Simple Measurement Mode: Cw')
            
                self.RfRst()
                self.Rcc1010.SmartIni('ExtTrigMux2')
                dCfg = varargin[1]

                if not ('fCenter' in dCfg):
                    print('RfMeas: fCenter not specified!')
                    traceback.print_exc()

                if not ('TMeas' in dCfg):
                    print('RfMeas: TMeas not specified!')
                    traceback.print_exc()

                if not ('NrFrms' in dCfg):
                    print('RfMeas: NrFrms not specified!')
                    traceback.print_exc()
                   
                if not ('N' in dCfg):
                    print('RfMeas: N not specified!')
                    traceback.print_exc()
                
                if not ('TInt' in dCfg):
                    print('RfMeas: TInt not specified!')
                    traceback.print_exc()
                

                dCfg = self.ChkMeasCfg(dCfg)
                

                self.Rf_fStrt = dCfg['fCenter']
                self.Rf_fStop = dCfg['fCenter']
                self.Rf_TRampUp = dCfg['TMeas']
                self.Rf_TRampDo = 0
                
                self.Rcc1010.DevProcCall("SetModRun")
                self.Rcc1010.DevProcCall("InitBasic")
                self.Rcc1010.DevProcCall("OutputStaticFrequency", dCfg['fCenter'])
                self.Rcc1010.DevProcCall("SetTxReg",1, int('11001111',2))
                self.Rcc1010.DevProcCall("SetTxReg",2, int('11001111',2))
                self.Rcc1010.DevProcCall("SetTxReg",3, int('11001111',2))
                self.Rcc1010.DevProcCall("StartCoarseTuneCalibration")
                self.Rcc1010.DevProcCall("StartFQMCalibration")    
                self.Rcc1010.DevProcCall("OutputStaticFrequency", dCfg['fCenter'])            
                
                # Calculate Sampling Rate
                N = dCfg['N']
                IniCic = 1
                if ('IniCic' in dCfg):
                    if dCfg['IniCic'] == 0:
                        IniCic = 0

                if IniCic > 0:
                    Ts = dCfg['TMeas']/N
                    fs = 1/Ts
                    fAdc = self.Get('fAdc') 
                    Div = floor(fAdc/fs)
                    if Div < 1:
                        Div = 1

                    if self.cDebugInf > 0: 
                        print('  fAdc:  ', (fAdc/1e6), ' MHz')
                        print('  CicR:  ', (Div), ' ')
                        print('  TSamp: ', (N/(fAdc/Div)/1e-6), ' us')  
                                        
                    # Configure CIC Filter 
                    self.CfgCicFilt(Div)

                self.Set('N', N)

                self.Fpga_MimoSeqRst()

                self.Rad_FrmCtrlCfg_RegChnCtrl = self.cFRMCTRL_REG_CH0CTRL_FRMCNTRENA + self.cFRMCTRL_REG_CH0CTRL_WAITENA + self.cFRMCTRL_REG_CH0CTRL_GLOBWAITENA    
                self.BrdSampIni()

                # Programm Timing to Board: Nios is used to initialize
                # pattern
                
                if dCfg['TInt'] < dCfg['TMeas'] + 1e-6:
                    dCfg['TInt'] = dCfg['TMeas'] + 1e-6
                    print('TInt to short; set to ', (dCfg['TInt']/1e-6), ' us')
                


                # TODO Set Timing 
                self.BrdSetTimCont('ContUnifM1', dCfg['TInt']);
                
                self.Setfs();
                self.SetFuSca(); 
                
                
                if ('Strt' in dCfg):
                    if dCfg['Strt'] > 0:
                        self.BrdAccessData()
                        self.BrdSampStrt()
                    
                else:
                    self.BrdAccessData()
                    self.BrdSampStrt()
             
                              
            elif stMod == 'RccMs':
                print('RCC being Master and Power Control with ')
                lRampCfg = varargin[1]
                dCfg = varargin[2] 
                self.RfRst()
                # Initialize Transceivers and Receivers
                self.RfRxEna()
                if 'TxPwr' in dCfg:
                    self.TxPwr = dCfg['TxPwr']
                self.RfTxPaEna(self.TxPwr)
                self.Rcc1010.SmartIni('ExtTrigMux2PaCtrl')
                
                if not ('NLoop' in dCfg):
                    dCfg['NLoop'] = 1
                
                dCfg = self.ChkMeasCfg(dCfg)

                if isinstance(lRampCfg,list):
                    self.SetActRampParams(dCfg)
                    # Configure Ramp Memory with P0 - P4 to set transmit antenna
                    # [P4,P3,P2,P1,P0] -> used to control Dpa signals
                    # 0: all Tx off
                    # 1: Tx1
                    # 2: Tx2 ...
                    
                    self.Rcc1010.ProgRampMemory(int('0x1000',0), lRampCfg)

                    nSet = len(lRampCfg)

                    TRamp = self.Rcc1010.GetRampSeqDuration(lRampCfg)
                    TInt = dCfg['TInt']
                    

                    if (TInt + 0.1e-3) < TRamp*dCfg['NLoop']:
                        TInt = TRamp*dCfg['NLoop'] + 0.1e-3
                        print('Increase Repition')

                    # Execute Processing Cals;
                    self.Rcc1010.DevProcCall("SetModRun")
                    self.Rcc1010.DevProcCall("InitBasic")
                    self.Rcc1010.DevProcCall("OutputStaticFrequency", (dCfg["fStrt"] + dCfg["fStop"])/2)
                    self.Rcc1010.DevProcCall("SetTxReg",1, int('11101111',2))
                    self.Rcc1010.DevProcCall("SetTxReg",2, int('11101111',2))
                    self.Rcc1010.DevProcCall("SetTxReg",3, int('11101111',2))
                    self.Rcc1010.DevProcCall("StartCoarseTuneCalibration")
                    self.Rcc1010.DevProcCall("StartFQMCalibration")

                    # Calculate Sampling Rate
                    N = dCfg['N']
                    IniCic = 1
                    if 'IniCic' in dCfg:
                        if dCfg['IniCic'] == 0:
                            IniCic = 0
                    
                    if IniCic > 0:
                        Ts = dCfg['TRampUp']/N
                        fs = 1/Ts
                        fAdc = self.Get('fAdc') 
                        Div = floor(fAdc/fs)
                        if Div < 1:
                            Div = 1
                        print('  fAdc:  ', (fAdc/1e6), ' MHz')
                        print('  CicR:  ', (Div), ' ')
                        print('  TSamp: ', (N/(fAdc/Div)/1e-6), ' us')                      
                        # Configure CIC Filter 
                        self.CfgCicFilt(Div)
                    self.Set('N', N)

                    self.Fpga_MimoSeqRst()
                    self.Fpga_MimoSeqNrEntries(1)
                    Regs = self.Rcc1010.GetRegsProcCallExtTrig(int('0x1000',0), nSet, dCfg['NLoop'], 0)
                    self.Fpga_MimoSeqSetCrocRegs(0, Regs)

                    if self.Gps > 0:
                        print("Enable Padding")
                        self.Rad_FrmCtrlCfg_RegChnCtrl = self.cFRMCTRL_REG_CH0CTRL_FRMCNTRENA + self.cFRMCTRL_REG_CH0CTRL_WAITENA + self.cFRMCTRL_REG_CH0CTRL_GLOBWAITENA + self.cFRMCTRL_REG_CH0CTRL_PADENA    
                    else:
                        self.Rad_FrmCtrlCfg_RegChnCtrl = self.cFRMCTRL_REG_CH0CTRL_FRMCNTRENA + self.cFRMCTRL_REG_CH0CTRL_WAITENA + self.cFRMCTRL_REG_CH0CTRL_GLOBWAITENA     
                    self.BrdSampIni()

                    # Use External events to synchronize boards to
                    # master timing
                    if (self.Rad_Role >= self.cRAD_ROLE_SL):
                        dCfg['IniEve'] = 1
                        dCfg['ExtEve'] = 1
                        print('Set Role Slave')
                
                    # Programm Timing to Board: Nios is used to initialize
                    # pattern
                    IniEve = 1
                    if 'IniEve' in dCfg:
                        IniEve = dCfg['IniEve']
                    IniTim = 0.2e-3
                    if 'IniTim' in dCfg:
                        IniTim = dCfg['IniTim']
                    CfgTim = 60e-6
                    if 'CfgTim' in dCfg:
                        CfgTim = dCfg['CfgTim']
                    if CfgTim < 30e-6:
                        CfgTim = 30e-6
                        print('RCC1010 configuration time set to 30 us')

                    MeasEve = 0
                    if 'ExtEve' in dCfg:
                        MeasEve = dCfg['ExtEve']
                    if  TInt < TRamp*dCfg['NLoop'] + CfgTim:
                        TInt = TRamp*dCfg['NLoop'] + CfgTim
                        print('TInt to short; increase to ', TInt, ' us')

                    dTimCfg = dict()

                    dTimCfg['MeasEveN'] = 1
                    if (self.Rad_Role == self.RAD_ROLE_SL):
                        TInt = TInt - 10e-6
                    elif (self.Rad_Role == self.RAD_ROLE_INTGPS):
                        if 'ExtEveNMeas' in dCfg:
                            dTimCfg['MeasEveN'] = dCfg['ExtEveNMeas']
                            dTimCfg['MeasTimAct'] = TRamp*dCfg['NLoop'] + CfgTim;  
                        
                    
                    dTimCfg['IniEve'] = IniEve                          #   Use ExtEve after Ini phase
                    dTimCfg['IniTim'] = IniTim                          #   Duration of Ini phase in us 
                    dTimCfg['CfgTim'] = CfgTim                          #   Configuration: Configure RCC for ExtTrig 
                    dTimCfg['MeasEve'] = MeasEve                         #   Use ExtEve after meas phase
                    dTimCfg['MeasTim'] = TInt - dTimCfg['CfgTim']       

                    # TODO Set Timing 
                    self.BrdSetTim_RccM1PW_Ms(dTimCfg)
                        
                    self.Setfs();
                    self.SetFuSca(); 
                    self.Computation.SetParam('Tp', dCfg['TInt']);
                    self.Computation.SetParam('Np', dCfg['NLoop']);
                    self.Computation.SetParam('fStrt',   dCfg['fStrt']);
                    self.Computation.SetParam('fStop',   dCfg['fStop']);
                    self.Computation.SetParam('TRampUp', dCfg['TRampUp']); 
                    
                    if 'Strt' in dCfg:
                        if dCfg['Strt'] > 0:
                            self.BrdAccessData()
                            self.BrdSampStrt()
                    else:
                        self.BrdAccessData()                        
                        self.BrdSampStrt()     

            else:
                print("Measurement Mode not known")

    # DOXYGEN ------------------------------------------------------
    #> @brief Display last chirp configuration
    #>    
    def     RfDispCfg(self):
        print('Rf Cfg')
        print('  fStrt:    ', (self.Rf_fStrt/1e9),' GHz')
        print('  fStop:    ', (self.Rf_fStop/1e9),' GHz')
        print('  TUp  :    ', (self.Rf_TRampUp/1e-6),' us')
        print('  TDown:    ', (self.Rf_TRampDo/1e-6),' us')

    def     RfRrn7745Ini(self, *varargin):
        if len(varargin) > 0:
            dCfg = varargin[0]
        else:
            dCfg = dict()
            dCfg['Mask'] = self.Rf_Rrn7745_Cfg1_Mask
            dCfg['Tst_I'] = self.Rf_Rrn7745_Cfg1_Tst_I
            dCfg['Tst_Q'] = self.Rf_Rrn7745_Cfg1_Tst_Q
            dCfg['Lo_I'] = self.Rf_Rrn7745_Cfg1_Lo_I
            dCfg['Lo_Q'] = self.Rf_Rrn7745_Cfg1_Lo_Q
            dCfg['Ena_Mix'] = self.Rf_Rrn7745_Cfg1_Ena_Mix
            dCfg['Mux'] = self.Rf_Rrn7745_Cfg1_Mux
            dCfg['Dc1'] = self.Rf_Rrn7745_Cfg1_Dc1
            dCfg['Dc2'] = self.Rf_Rrn7745_Cfg1_Dc2
            dCfg['Dc3'] = self.Rf_Rrn7745_Cfg1_Dc3
            dCfg['Dc4'] = self.Rf_Rrn7745_Cfg1_Dc4             
        if self.cDebugInf > 1:
            print('IFRX Rf_Init')
        self.SetRrn7745(dCfg)
    
    def     RfRrn7745GetReg(self, Mask, Adr):
        Adr = Adr % 2**14
        Val = [];
        if DebugInf > 1:
            print('IFRX Regs Ini')
        Val = self.Fpga_GetIfrxReg(Mask, Adr)
        return Ret
        

    def     RfRst(self):
        self.Rcc1010.DevRst()
    
   

    def     RfRpn7720Ini(self, *varargin):
        if len(varargin) > 0:
            dCfg = varargin[0]
        else:
            dCfg['Mask'] = self.Rf_Rpn7720Cfg1_Mask
            dCfg['LoChain'] = self.Rf_Rpn7720Cfg1_LoChain
            dCfg['Pa1'] = self.Rf_Rpn7720Cfg1_Pa1
            dCfg['Pa2'] = self.Rf_Rpn7720Cfg1_Pa2
            dCfg['PaCon'] = self.Rf_Rpn7720Cfg1_PaCon
            dCfg['Mux'] = self.Rf_Rpn7720Cfg1_Mux                 
        if self.cDebugInf > 1:
            print('IFTX Rf_Init')
        self.SetRpn7720(dCfg)

    def     SetRpn7720(self, dCfg):
        Data = []
        # Programm LoChain
        Data = Data + [ (2**15) + (2**14) + 0*(2**8) + (dCfg['LoChain'] % 256)]              
        # Programm Pa1
        Data = Data + [ (2**15) + (2**14) + 1*(2**8) + (dCfg['Pa1'] % 256)]
        # Programm Pa2
        Data = Data + [ (2**15) + (2**14) + 2*(2**8) + (dCfg['Pa2'] % 256)]
        # Programm PaCon
        Data = Data + [ (2**15) + (2**14) + 4*(2**8) + (dCfg['PaCon'] % 256)]
        # Programm Mux
        Data = Data + [ (2**15) + (2**14) + 5*(2**8) + (dCfg['Mux'] % 256)]
        Ret = self.Fpga_SetIfTxReg(dCfg['Mask'], asarray(uint32(Data)))            
        return Ret

    def     Fpga_SetPaCtrl(self, Mask, Strt, Reg):
        if len(Reg) > 28:
            Reg = Reg[0:28]
        Cod = int('0x9215',0)
        FpgaCmd = asarray([Mask, 1 , Strt])
        FpgaCmd = append(FpgaCmd, Reg)
        Ret = self.CmdSend(0,Cod,FpgaCmd)
        Ret = self.CmdRecv()
        return Ret

    def     Fpga_GetDpaReg(self, Mask, Reg):
        Cod = int('0x9205',0)
        FpgaCmd = asarray([Mask, 2 , Reg])
        Ret = self.CmdSend(0,Cod,FpgaCmd)
        Ret = self.CmdRecv()
        return Ret


    def     Fpga_SetIfTxReg(self, Mask, Reg):
        Cod = int('0x9205',0)
        FpgaCmd = asarray([Mask, 1])
        FpgaCmd = append(FpgaCmd, asarray(Reg))
        Ret = self.CmdSend(0, Cod, FpgaCmd)
        Ret = self.CmdRecv()
        return Ret

    def     Fpga_GetIfTxReg(self, Mask, Reg):
        Cod = int('0x9205',0)
        FpgaCmd = asarray([Mask, 2])
        FpgaCmd = append(FpgaCmd, asarray(Reg))        
        Ret = self.CmdSend(0, Cod, FpgaCmd)
        Ret = self.CmdRecv()
        return Ret

    def     Fpga_SetIfrxReg(self, Mask, Reg):
        Cod = int('0x9206',0)
        FpgaCmd = asarray([Mask, 1]) 
        FpgaCmd = append(FpgaCmd, asarray(Reg)) 
        Ret = self.CmdSend(0, Cod, FpgaCmd)
        Ret = self.CmdRecv()
        return Ret

    def     Fpga_GetIfrxReg(self, Mask, Reg):
        Cod = int('0x9206',0)
        FpgaCmd = asarray([Mask, 2])
        FpgaCmd = append(FpgaCmd, asarray(Reg))
        Ret = self.CmdSend(0,Cod,FpgaCmd)
        Ret = self.CmdRecv()
        return Ret

    def     SetRrn7745(self, dCfg):
        # Programm Tst_I
        Data = [(2**15) + (2**14) + 32*(2**8) + (dCfg['Tst_I'] % 256)]
        # Programm Tst_Q
        Data.append( (2**15) + (2**14) + 33*(2**8) + (dCfg['Tst_Q'] % 256))
        # Programm Lo_I
        Data.append( (2**15) + (2**14) + 34*(2**8) + (dCfg['Lo_I'] % 256))               
        # Programm Lo_Q
        Data.append( (2**15) + (2**14) + 35*(2**8) + (dCfg['Lo_Q'] % 256))               
        # Programm Ena_Mix
        Data.append( (2**15) + (2**14) + 36*(2**8) + (dCfg['Ena_Mix'] % 256))            
        # Programm Mux
        Data.append( (2**15) + (2**14) + 37*(2**8) + (dCfg['Mux'] % 256))               
        # Programm DC1
        Data.append( (2**15) + (2**14) + 40*(2**8) + (dCfg['Dc1'] % 256))            
        # Programm DC2
        Data.append( (2**15) + (2**14) + 41*(2**8) + (dCfg['Dc2'] % 256))                
        # Programm DC3
        Data.append( (2**15) + (2**14) + 42*(2**8) + (dCfg['Dc3'] % 256))           
        # Programm DC4
        Data.append( (2**15) + (2**14) + 43*(2**8) + (dCfg['Dc4'] % 256))              
        Ret = self.Fpga_SetIfrxReg(dCfg['Mask'], asarray(uint32(Data)))
        return Ret
    
    def     RfRcc1010_GetRegsRampScenarioExtTrig(self, Adr, NSet, NLoop, ExitCfg):
        if Adr < int('0x1000',0):
            Adr = uint32('0x1000',0)
        if Adr > int('0x7FF0',0):
            Adr = uint32('0x1000',0)
        NSet = uint32(NSet % 1792)
        NLoop = uint32(NLoop % (2**16))
        ExitCfg = uint32(ExitCfg % (2**16))

        Data = asarray([Adr, NSet, NLoop])
        Regs = append(self.RCC1010_CallCode_StartRampScenarioExtTrig, Data)
        Regs = append(Regs, ExitCfg)     
        Regs = Regs + self.RCC1010_Reg_ProcessingCall_Adr*(2**16) + (2**31) 
        return Regs
                
    # DOXYGEN ------------------------------------------------------
    #> @brief Configure CIC filter for clock divider
    #>
    #> Configure CIC filter: Filter transfer function is adjusted to sampling rate divider
    #> 
    #> @param[in] Div: 
    def     CfgCicFilt(self, Div):
        if Div >= 16:
            self.Set('CicEna')                         
            self.Set('CicR', Div)
            self.Set('CicDelay', 16)
        elif Div >= 8:
            self.Set('CicEna')                        
            self.Set('CicR',Div)
            self.Set('CicDelay',8)
        elif Div >= 4:
            self.Set('CicEna')                         
            self.Set('CicR',Div)
            self.Set('CicDelay',4)
        elif Div >= 2:    
            self.Set('CicEna')                           
            self.Set('CicR',Div)
            self.Set('CicDelay',2)
        else:
            self.Set('CicDi')       
  
    # DOXYGEN ------------------------------------------------------
    #> @brief Check measurement configuration 
    #> 
    #> Check measurement configuration structure
    #> 
    #> @param[in] Cfg: 
    def     ChkMeasCfg(self, dCfg):
        if 'IniEve' in dCfg:
            IniEve = floor(dCfg['IniEve'])
            IniEve = (IniEve % 2)
            if IniEve != dCfg['IniEve']:
                print('Rbk2IfxTx2Rx16: IniEve set to ', (IniEve))
            dCfg['IniEve'] = IniEve
        
        if 'ExtEve' in dCfg:
            ExtEve = floor(dCfg['ExtEve'])
            ExtEve = (ExtEve % 2)
            if ExtEve != dCfg['ExtEve']:
                print('Rbk2IfxTx2Rx16: ExtEve set to ', (ExtEve))
            dCfg['ExtEve'] = ExtEve
                      
        if 'N' in dCfg:
            N = ceil(dCfg['N']/8)*8
            if (N != dCfg['N']):
                print('Rbk2IfxTx2Rx16: N must be a mulitple of 8 -> Set N to ', N)
            if N > 4096:
                N = 4096
                print('Rbk2IfxTx2Rx16: N to large -> Set N to ', N)
            dCfg['N'] = N
            if  N < 8:
                N = 8
                print('Rbk2IfxTx2Rx16: N to small -> Set N to ', N)

        # Check number of repetitions: standard timing modes can only can be used with 256 repttions
        # Np must be greater or equal to 1 and less or equal than 256
        if 'Np' in dCfg:
            Np = ceil(dCfg['Np'])
            if Np < 1: 
                Np = 1
            if Np > 256:
                Np = 256
            if Np != dCfg['Np']:
                print('Rbk2IfxTx2Rx16: Np -> Set Np to ', Np)
            dCfg['Np'] = Np

        # Check number of frames: at least one frame must be measured
        #if 'NrFrms' in dCfg:
        #    dCfg['NrFrms'] = ceil(dCfg['NrFrms'])
        #    if dCfg['NrFrms'] < 1:
        #        dCfg['NrFrms'] = 1
        #        print('Rbk2IfxTx2Rx16: NrFrms < 1 -> Set to 1')
        
        # Initialization time (used to generate dummy frame and reset signal processing chain and data interface)
        if 'IniTim' in dCfg:
            IniEve = 1
            if 'IniEve' in dCfg:
                IniEve = dCfg['IniEve']
            # If external event is programmed after init, a too long ini time can result that the event is missed
            if (IniEve > 0) and (dCfg['IniTim'] > 5e-3):
                print('Rbk2IfxTx2Rx16: Trigger Event could be missed (Meas does not start)')

        # Tx Field: SeqTrig MMP < 2.0.0 support 16 entries in the sequence table:
        if 'TxSeq' in dCfg:
            TxSeq = dCfg['TxSeq']
            NSeq = len(TxSeq)
            if NSeq > 16:
                print('Rbk2IfxTx2Rx16: TxSeq to long -> limit to 16')    
                TxSeq = TxSeq[1:16]
            if NSeq < 1:
                TxSeq = 0
                print('Rbk2IfxTx2Rx16: TxSeq empty -> Set to 0')  
            dCfg['TxSeq'] = TxSeq

        return dCfg

    def     BrdSetTim_RccM1PW(self, dCfg):
        fSeqTrig = self.Rad_fAdc;
        Seq = SeqTrig.SeqTrig(fSeqTrig)

        print('BrdSetTim_RccM1PW')
        SeqTrigCfg = dict()
        SeqTrigCfg['Mask'] = 1;
        SeqTrigCfg['Ctrl'] = Seq.SEQTRIG_REG_CTRL_IRQ2ENA; # Enable interrupt event on channel 2 
        SeqTrigCfg['ChnEna'] = Seq.SEQTRIG_REG_CTRL_CH0ENA + Seq.SEQTRIG_REG_CTRL_CH1ENA;
        SeqTrigCfg['Seq'] = list()

        # Phase 0: Ini with dummy frame: (lSeq, 'RccCfg', TCfg, Adr, TxChn(Idx)-1);
        if dCfg['IniEve'] > 0:
            lSeq = Seq.IniSeq('IniExt', dCfg['IniTim'])    
        else:
            lSeq = Seq.IniSeq('Ini', dCfg['IniTim'])     
        # Phase 1: Cfg: Generate Irq:
        lSeq = Seq.AddSeq(lSeq, 'RccCfg', dCfg['CfgTim'], 2, 0)
        # Phase 2: Meas 
        if dCfg['MeasEve'] > 0:
            lSeq = Seq.AddSeq(lSeq,'RccMeasWait', dCfg['MeasTim'], 1, 0)
        else:
            lSeq = Seq.AddSeq(lSeq,'RccMeas', dCfg['MeasTim'], 1, 0)
        SeqTrigCfg['Seq'] = lSeq

        self.SeqCfg = SeqTrigCfg

        self.Fpga_SeqTrigRst(self.SeqCfg['Mask'])
        NSeq = len(self.SeqCfg['Seq'])
        print('Run SeqTrig: ', NSeq, ' Entries')
    
        for Idx in range(0,NSeq):
            self.Fpga_SeqTrigCfgSeq(self.SeqCfg['Mask'],Idx, self.SeqCfg['Seq'][Idx])


    def     BrdSetTim_RccM1PW_Ms(self, dCfg):
        fSeqTrig = self.Rad_fAdc;
        Seq = SeqTrig.SeqTrig(fSeqTrig)
        
        print('BrdSetTim_RccM1PW_Ms')
        SeqTrigCfg = dict()
        SeqTrigCfg['Mask'] = 1
        SeqTrigCfg['Ctrl'] = Seq.SEQTRIG_REG_CTRL_IRQ2ENA  # Enable interrupt event on channel 2 
        SeqTrigCfg['ChnEna'] = Seq.SEQTRIG_REG_CTRL_CH0ENA + Seq.SEQTRIG_REG_CTRL_CH1ENA + Seq.SEQTRIG_REG_CTRL_CH3ENA
        SeqTrigCfg['Seq'] = list()

        if dCfg['MeasEveN'] > 1:
            # in this mode the board is configured as slave
            if dCfg['IniEve'] > 0:
                lSeq = Seq.IniSeq('Ini', dCfg['IniTim'] - 10e-6)
                lSeq = Seq.AddSeq(lSeq, 'WaitEve', 10e-6, 2, 0)
                lSeq[-1]['Chn3Cfg'] = 2**30
            else:
                lSeq = Seq.IniSeq('Ini', dCfg['IniTim'] - 10e-6)
                lSeq = Seq.AddSeq(lSeq, 'Wait', 10e-6, 2, 0)
                lSeq[-1]['Chn3Cfg'] = 2**30
            
            # Phase 1: Cfg: Generate Irq:
            lSeq = Seq.AddSeq(lSeq, 'RccCfg', dCfg['CfgTim'], 3, 0)                                 #2
            # Set input trigger to external input event
            lSeq[-1]['Chn3Cfg'] = 2**30 + 2**31 
            lSeq = Seq.AddSeq(lSeq,'RccMeasN', dCfg['MeasTim'], 2, 4, dCfg['MeasEveN'] - 1, 0)      #3
            lSeq[-1]['Chn3Cfg'] = 2**30 + 2**31 
            lSeq = Seq.AddSeq(lSeq, 'RccCfg', dCfg['CfgTim'], 5, 0)                                 #4
            lSeq[-1]['Chn3Cfg'] = 2**30 + 2**31
            lSeq = Seq.AddSeq(lSeq,'RccMeasWait', dCfg['MeasTimAct'], 2, 0)                         #5
            lSeq[-1]['Chn3Cfg'] = 2**30 + 2**31

        else:
            # Phase 0: Ini with dummy frame: (lSeq, 'RccCfg', TCfg, Adr, TxChn(Idx)-1);
            if dCfg['IniEve'] > 0:
                lSeq = Seq.IniSeq('Ini', dCfg['IniTim'] - 10e-6)
                lSeq = Seq.AddSeq(lSeq, 'WaitEve', 10e-6, 2, 0)
                if self.Rad_Role >= self.cRAD_ROLE_SL:
                    # Set input trigger to external input event
                    lSeq[-1]['Chn3Cfg'] = 2**30             
            else:
                lSeq = Seq.IniSeq('Ini', dCfg['IniTim'] - 10e-6) 
                lSeq = Seq.AddSeq(lSeq, 'Wait', 10e-6, 2, 0)
                if self.Rad_Role >= self.cRAD_ROLE_SL:
                    # Set input trigger to external input event
                    lSeq[-1]['Chn3Cfg'] = 2**30

            # Phase 1: Cfg: Generate Irq:
            lSeq = Seq.AddSeq(lSeq, 'RccCfg', dCfg['CfgTim'], 3, 0)
            if self.Rad_Role >= self.cRAD_ROLE_SL:
                # Set input trigger to external input event
                lSeq[-1]['Chn3Cfg'] = 2**30 + 2**31  
            else:
                lSeq[-1]['Chn3Cfg'] = 2**31

            # Phase 2: Meas 
            if dCfg['MeasEve'] > 0:
                lSeq = Seq.AddSeq(lSeq,'RccMeasWait', dCfg['MeasTim'], 2, 0)
            else:
                lSeq = Seq.AddSeq(lSeq,'RccMeas', dCfg['MeasTim'], 2, 0)     
            if self.Rad_Role >= self.cRAD_ROLE_SL:
                # Set input trigger to external input event
                lSeq[-1]['Chn3Cfg'] = 2**30 + 2**31  
            else:
                lSeq[-1]['Chn3Cfg'] = 2**31


        SeqTrigCfg['Seq'] = lSeq

        self.SeqCfg = SeqTrigCfg

        self.Fpga_SeqTrigRst(self.SeqCfg['Mask'])
        NSeq = len(self.SeqCfg['Seq'])
        print('Run SeqTrig: ', (NSeq), ' Entries')
        for Idx in range(0,NSeq):
             self.Fpga_SeqTrigCfgSeq(self.SeqCfg['Mask'], Idx, self.SeqCfg['Seq'][Idx])









    def     BrdSetTim_RccM1PW_MsIniStop(self, dCfg):
        fSeqTrig = self.Rad_fAdc;
        Seq = SeqTrig.SeqTrig(fSeqTrig)
        print('BrdSetTim_RccM1PW_MsIniStop')
        SeqTrigCfg = dict()
        SeqTrigCfg['Mask'] = 1
        SeqTrigCfg['Ctrl'] = Seq.SEQTRIG_REG_CTRL_IRQ2ENA                                        # Enable interrupt event on channel 2 
        SeqTrigCfg['ChnEna'] = Seq.SEQTRIG_REG_CTRL_CH0ENA + Seq.SEQTRIG_REG_CTRL_CH1ENA
        SeqTrigCfg['Seq'] = list()

        # Phase 0: Ini with dummy frame: (lSeq, 'RccCfg', TCfg, Adr, TxChn(Idx)-1);
        if dCfg['IniEve'] > 0:
            lSeq = Seq.IniSeq('Ini', dCfg['IniTim'] - 10e-6)
            lSeq = Seq.AddSeq(lSeq, 'WaitEve', 10e-6, 2, 0)
        else:
            lSeq = Seq.IniSeq('Ini', dCfg['IniTim'] - 10e-6) 
            lSeq = Seq.AddSeq(lSeq, 'Wait', 10e-6, 2, 0)    
        # Phase 1: Cfg: Generate Irq:
        lSeq = Seq.AddSeq(lSeq, 'RccCfg', dCfg['CfgTim'], 3, 0)
        lSeq[2]['Chn3Cfg'] = (2**31)
        # Phase 2: Meas
        # stay in this phase for ever loop back to itselfe 
        if dCfg['MeasEve'] > 0:
            lSeq = Seq.AddSeq(lSeq,'RccMeasWait', dCfg['MeasTim'], 4, 0)
        else:
            lSeq = Seq.AddSeq(lSeq,'RccMeas', dCfg['MeasTim'], 4, 0)     
        lSeq[3]['Chn3Cfg'] = (2**31)
        if dCfg['MeasEve'] > 0:
            lSeq = Seq.AddSeq(lSeq,'RccMeasWait', dCfg['MeasTim'], 4, 0)
        else:
            lSeq = Seq.AddSeq(lSeq,'RccMeas', dCfg['MeasTim'], 4, 0)     
        lSeq[4]['Chn3Cfg'] = (2**31)        

        SeqTrigCfg['Seq'] = lSeq

        self.SeqCfg = SeqTrigCfg

        self.Fpga_SeqTrigRst(self.SeqCfg['Mask'])
        NSeq = len(self.SeqCfg['Seq'])
        print('Run SeqTrig: ', (NSeq), ' Entries')
        for Idx in range(0,NSeq):
             self.Fpga_SeqTrigCfgSeq(self.SeqCfg['Mask'], Idx, self.SeqCfg['Seq'][Idx])




#< @file        DevDriver.py                                                                
#< @author      Haderer Andreas (HaAn)                                                  
#< @date        2013-06-13          
#< @brief       Class for device driver configuration functions
#< @version     1.0.1
from    numpy import *

class   DevDriver(object):
    
    def __init__(self):
        self.lRegs      =   list()
        pass

    # DOXYGEN ------------------------------------------------------
    #> @brief Generate register values with class predefined flags
    #>  
    #> This function can be used to generate register values. The function automatically ensures that reserver bits are preserved.
    #>
    #>  @param[in]  AdrName: if value is a number than it is interpreted as address; if the value is a string, then as the register name
    #>
    #>  @param[in]  Ini value of the register
    #>  
    #>  @param[in]  variable number of flags and values for the flags: e.g. , 'Div', 1: if the divider flag exists
    #>
    #>  @return     Value of the register
    #>       
    #>
    #> e.g. Generate register values with flags
    #>   @code
    #>   Brd        =   Radarbook('PNet','192.168.1.1')
    #>   Adf5904    =   DevAdf5904(Brd,USpiCfg)
    #>   RegR0      =   Adf5904.GenRegFlg('R0', 0, 'PupLo', 1)
    #>   @endcode          
    def     GenRegFlag(self, AdrName, Ini, *varargin):
        RegVal      =   -1

        if isinstance(AdrName, str):
            RegIdx      =   self.FindRegByName(AdrName) 
            NArg        =   len(varargin)
            ArgIdx      =   0
            RegVal      =   Ini
            while ArgIdx < NArg:
                RemArg  =   ArgIdx
                if (RemArg % 2) == 0:
                    stField     =   varargin[ArgIdx]
                    Val         =   varargin[ArgIdx+1]    
                    RegVal      =   self.GenRegSetField(RegVal, RegIdx, stField, Val)    
                    ArgIdx      = ArgIdx + 2
                else:
                    # cancel initialization
                    ArgIdx      = NArg
        
            RegVal      =   self.GenRegSetDefault(RegVal, RegIdx);
        else:
            pass

        return  RegVal    

        
    # DOXYGEN ------------------------------------------------------
    #> @brief Print register map in Matlab command window
    #>  
    #> Print register map in the command window of Matlab. This function can be used to find the names of the registers.
    #> 
    #> e.g. After calling the function the register map is printed 
    #> @code
    #> Register Name:  R0 @Adr: 0
    #> Register Name:  R1 @Adr: 1
    #> Register Name:  R2 @Adr: 2
    #> Register Name:  R3 @Adr: 3
    #> Register Name:  R4 @Adr: 4
    #> Register Name:  R5 @Adr: 5
    #> Register Name:  R6 @Adr: 6
    #> Register Name:  R7 @Adr: 7
    #> Register Name:  R8 @Adr: 8
    #> Register Name:  R9 @Adr: 9
    #> Register Name:  R10 @Adr: 10
    #> @endcode             
    def     PrntRegMap(self, *varargin):
        print("[DevDriver]: PrntRegMap")        
        for dElem in self.lRegs:
            stOut   =   'Register Name:  ' +  dElem["Name"] + ' @Adr: ' + str(dElem["Adr"])  
            print(stOut)
         
        
    # DOXYGEN ------------------------------------------------------
    #> @brief Print flags of registers in Matlab command window
    #>  
    #> This function can be used to find the names of the flags.
    #>   
    #> @param[in]   Name of register (optional)
    #>
    #> After calling the function with argument 'R0' the flags for R0 are printed in the command window
    #>  @code
    #> Reg: R0 @Adr: 0
    #>  Field: Ctrl
    #>  Field: Res
    #>  Field: DoutVSel
    #>  Field: LoPinBias
    #>  Field: PupLo
    #>  Field: PupChn1
    #>  Field: PupChn2
    #>  Field: PupChn3
    #>  Field: PupChn4
    #>  Field: Res
    #>  @endcode
    def     PrntRegFlags(self, *varargin):
        if len(varargin) > 0:
            if isinstance(varagin[0],str):
                RegIdx      =   obj.FindRegByName(varargin[0]);
                dReg        =   self.lRegs[RegIdx]
                stOut       =   'Reg: ' + dReg["Name"] + ' @Adr: ' + str(dReg["Adr"])
                print(stOut)
                for dField in dReg["lFields"]:
                    print('  Field: ',dField["Name"])
        else:
            for dElem in self.lRegs:
                stOut       =   'Reg: ' + dElem["Name"] + ' @Adr: ' + str(dElem["Adr"])
                print(stOut)                
                for dField in dElem["lFields"]:
                    print('  Field: ',dField["Name"])



    #   @function       GenRegSetField                                                             
    #   @author         Haderer Andreas (HaAn)                                                  
    #   @date           2015-07-27          
    #   @brief          Set Field of Register
    #   @param[in]      RegVal: register value
    #   @param[in]      RegIdx: Index of register in caRegs array
    #   @param[in]      stField: Name of field
    #   @param[in]      Value of field
    #   @return         New register value
    def     GenRegSetField(self, RegVal, RegIdx, stField, Val):
        dReg        =   self.lRegs[RegIdx]
        lFields     =   dReg["lFields"]
        RegVal      =   RegVal
        Ini         =   False
        for dEntry in lFields: 
            if dEntry["Name"] == stField:
                Ini         =   True
                Low         =   2**dEntry["Strt"]
                High        =   2**(dEntry["Stop"] + 1)
                if High < Low:
                    print('Field definition error')
                else:
                    Lim         =   2**(dEntry["Stop"] + 1 - dEntry["Strt"])
                    RegValLow   =   RegVal % Low
                    RegValHigh  =   floor(RegVal/High)
                    Val         =   Val % Lim
                    RegVal      =   RegValLow + Val*Low + RegValHigh*High;
        if Ini == False:
            print("Field not known: ", stField);
        
        return  RegVal

    def     FindRegByAdr(self, Adr):
        Idx         =   -1
        RegIdx      =   Idx
        for dReg    in self.lRegs:
            Idx     =   Idx + 1
            if dReg["Adr"] == Adr:
                RegIdx      =   Idx

        return RegIdx

   
    def     FindRegByName(self, stName):
        Idx         =   -1
        RegIdx      =   Idx
        for dReg    in  self.lRegs:
            Idx     =   Idx + 1
            if dReg["Name"] == stName:
                RegIdx      =   Idx
        return RegIdx

    #   @function       GenRegSetDefault                                                             
    #   @author         Haderer Andreas (HaAn)                                                  
    #   @date           2015-07-27          
    #   @brief          Set default value: ensure res and adr field
    #   values
    #   @param[in]      RegVal: register value
    #   @param[in]      RegIdx: Index of register in caRegs array 
    def     GenRegSetDefault(self, RegVal, RegIdx):
        dReg        =   self.lRegs[RegIdx]
        lFields     =   dReg["lFields"]
        RegVal      =   RegVal;
        for dField in lFields:
            if dField["Name"] == 'Res':
                Low         =   2**dField["Strt"]
                High        =   2**(dField["Stop"] + 1)
                Lim         =   2**(dField["Stop"] + 1 - dField["Strt"])
                RegValLow   =   RegVal % Low
                RegValHigh  =   floor(RegVal/High)
                Val         =   dField["Val"] % Lim
                RegVal      =   RegValLow + Val*Low + RegValHigh*High
            if dField["Name"] == 'Ctrl':
                Low         =   2**dField["Strt"]
                High        =   2**(dField["Stop"] + 1)
                Lim         =   2**(dField["Stop"] + 1 - dField["Strt"])
                RegValLow   =   RegVal % Low
                RegValHigh  =   floor(RegVal/High)
                Val         =   dField["Val"] % Lim
                RegVal      =   RegValLow + Val*Low + RegValHigh*High

        return RegVal

        





import  os
import  platform
import  sys
import  struct
from    numpy import *

__version__     =   "1.0.0"

class   DevRrn7745(object):

    def __init__(self, Rad, Mask):

        self.Rad                =   Rad
        self.Mask               =   Mask

        # Define Constants
        self.DefineConst()

        # Define Default register values
        self.RegTstI                        =   int("01111111",2)
        self.RegTstQ                        =   0
        self.RegLoI                         =   0
        self.RegLoQ                         =   0
        self.RegEnaMix                      =   0
        self.RegMux                         =   int("00010000",2)
        self.RegDc1                         =   0
        self.RegDc2                         =   0
        self.RegDc3                         =   0
        self.RegDc4                         =   0
        self.DevSetCfg()

    def SetCfg(self, dCfg):
        #   @function       SetCfg                                                             
        #   @author         Haderer Andreas (HaAn)                                                  
        #   @date           2015-07-27          
        #   @brief          Set Device Configuration
        #   @param[in]      dCfg
        #                       TxChn:      Transmit channel
        #                       TxPwr:      Transmit clocks
        #                       RxEna:      Receive Enable (0-15) bit 0 Rx 0 etc
        if "TxChn" in dCfg:
            TxChn   =   int(dCfg["TxChn"])
            if TxChn < 0:
              TxChn         =     0
            if TxChn > 2:
              TxChn         =     2
            
            if "TxPwr" in dCfg:
                TxPwr   =   int(dCfg["TxPwr"])                
                if TxPwr < 0:
                    TxPwr       =     0
                if TxPwr > 63:
                    TxPwr       =     63
            
            if TxChn == 0:
                self.RegPa1     =   0
                self.RegPa2     =   0
                self.RegPaCon   =   0
            elif TxChn  == 1:
                self.RegPa1     =   TxPwr
                self.RegPa2     =   0
                self.RegPaCon   =   self.cReg_PaCon_MuxP1
            elif TxChn == 2:
                self.RegPa1     =   0
                self.RegPa2     =   TxPwr
                self.RegPaCon   =   self.cReg_PaCon_MuxP2                            
            else:
                self.RegPa1     =   0
                self.RegPa2     =   0
                self.RegPaCon   =   0

        if "TxMask" in dCfg:
            TxMask  =   int(dCfg["TxMask"])
            if "TxPwr" in dCfg:
                TxPwr   =   int(dCfg["TxPwr"])                
                if TxPwr < 0:
                    TxPwr       =     0
                if TxPwr > 63:
                    TxPwr       =     63
            
            self.RegPaCon       =   0
            if (TxMask & int(1)) > 0:
                self.RegPa1     =   TxPwr
                self.RegPaCon   =   self.RegPaCon + self.cReg_PaCon_MuxP1
            if (TxMask & int(2)) > 0:    
                self.RegPa1     =   TxPwr
                self.RegPaCon   =   self.RegPaCon + self.cReg_PaCon_MuxP2             


        if (int(self.Rad.DebugInf) & int('0xF0',0)) > 64:
            print("RPN7720 Registers:")
            print(" LoChain %d:", self.RegLoChain)
            print(" Pa1 %d:", self.RegPa1)
            print(" Pa2 %d:", self.RegPa2)
            print(" PaCon %d:", self.RegPaCon)
            print(" Mux %d:", self.RegMux)

    def SetRegCfg(self, dCfg):
        #   @function       SetCfg                                                             
        #   @author         Haderer Andreas (HaAn)                                                  
        #   @date           2015-07-27          
        #   @brief          Set Device Configuration
        #   @param[in]      dCfg
        #                       LoChain:    LO Chain
        #                       Pa1:        Pa1
        #                       Pa2:        Pa2 
        #                       PaCon:      PaControl
        #                       Mux:        Mux                     
        if "LoChain" in dCfg:
            self.RegLoChain     =   dCfg["LoChain"]                     
        if "Pa1" in dCfg:
            self.RegPa1         =   dCfg["Pa1"]
        if "Pa2" in dCfg:
            self.RegPa2         =   dCfg["Pa2"]
        if "PaCon" in dCfg:
            self.RegPaCon       =   dCfg["PaCon"]
        if "Mux" in dCfg:
            self.RegMux         =   dCfg["Mux"] 
        
        if (int(self.Rad.DebugInf) & int('0xF0',0)) > 64:
            print("RPN7720 Registers:")
            print(" LoChain %d:", self.RegLoChain)
            print(" Pa1 %d:", self.RegPa1)
            print(" Pa2 %d:", self.RegPa2)
            print(" PaCon %d:", self.RegPaCon)
            print(" Mux %d:", self.RegMux)

    def DevSetCfg(self):
        # reset this function is currently not implemented
        pass

    def DevRst(self):
        # reset currently not implemented
        pass
    
    def DevEna(self):
        # enable currently not implemented
        pass

    def DevDi(self):
        # disable currently not implemented
        pass

    def DevSetReg(self, Regs):
        #   @function       SetCfg                                                             
        #   @author         Haderer Andreas (HaAn)                                                  
        #   @date           2015-07-27          
        #   @brief          Set Device Configuration
        #   @param[in]      Regs: register values beeing programmed over spi
        if (int(self.Rad.DebugInf) & int('0xF0',0)) > 16: 
            print("RPN7720 DevSetReg")
        NReg            =   shape(Regs)
        NEntr           =   int(NReg[0])
        Data            =   zeros(NEntr + 2,dtype='uint32')
        Data[0]         =   self.Mask                       # Mask for device 
        Data[1]         =   1
        Data[2:]        =   Regs               
        CmdCod          =   int("0x9205", 0)
        RxData          =   self.Rad.CmdSend(0, CmdCod, Data)
        Ret             =   self.Rad.CmdRecv()
        if Ret[0] == False:
            print("Error RPN7720 DevSetReg")
        return Ret
    
    def DevGetReg(self, Regs):
        pass

    def IniCfg(self, dCfg):
        #   @function       Ini                                                             
        #   @author         Haderer Andreas (HaAn)                                                  
        #   @date           2015-07-27          
        #   @brief          Programm initialization to device          
        self.SetCfg(dCfg)
        Ret     =   self.Ini();
        if Ret[0] == False:
            print("Error RPN7720: IniCfg")
        return Ret[0]

    def Ini(self):
        #   @function       Ini                                                             
        #   @author         Haderer Andreas (HaAn)                                                  
        #   @date           2015-07-27          
        #   @brief          Initialize device configure registers   
        Data            =   self.GenRegs()
        Ret             =   self.DevSetReg(Data)
        return Ret[0]

    def GenRegs(self):
        #   @function       GenRegs                                                             
        #   @author         Haderer Andreas (HaAn)                                                  
        #   @date           2015-07-27          
        #   @brief          Generate registers for programming      
        Data            =   zeros(5,dtype='uint32')
        Val             =   self.cWriteVerify + self.cReg_LoChain_Adr*2**8 + self.RegLoChain
        Data[0]         =   Val
        Val             =   self.cWriteVerify + self.cReg_Pa1_Adr*2**8 + self.RegPa1
        Data[1]         =   Val
        Val             =   self.cWriteVerify + self.cReg_Pa2_Adr*2**8 + self.RegPa2
        Data[2]         =   Val
        Val             =   self.cWriteVerify + self.cReg_PaCon_Adr*2**8 + self.RegPaCon
        Data[3]         =   Val
        Val             =   self.cWriteVerify + self.cReg_Mux_Adr*2**8 + self.RegMux
        Data[4]         =   Val
        return Data

    def DefineConst(self):
        self.cReg_LoChain_Adr           =   int("0x0", 0)
        self.cReg_Pa1_Adr               =   int("0x1", 0)
        self.cReg_Pa2_Adr               =   int("0x2", 0)
        self.cReg_PaCon_Adr             =   int("0x4", 0)
        self.cReg_Mux_Adr               =   int("0x5", 0)
        self.cWriteVerify               =   2**15 + 2**14
        self.cReg_PaCon_MuxP1           =   3
        self.cReg_PaCon_MuxP2           =   3*16  



from setuptools import setup, find_packages

setup(
    name='pyavlss2025',
    version='0.1.2',
    description='LSS-2025',
    author='Alessandro Vaccari',
    author_email='alessandro.vaccari@unife.it',
    packages=find_packages(),
    python_requires='>=3.6',
)